# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
Classes representing collections and libraries

A collection is a database of tracks. It is based on :class:`TrackDB` but has
the ability to be linked with libraries.

A library finds tracks in a specified directory and adds them to an associated
collection.
"""

from collections import deque
import logging
import threading
from typing import Deque, Dict, Iterable, List, MutableSequence, Optional, Set, Tuple

from gi.repository import (
    GLib,
    GObject,
    Gio,
)

from xl import common, event, settings, trax

logger = logging.getLogger(__name__)

COLLECTIONS: Set['Collection'] = set()


def get_collection_by_loc(loc: str) -> Optional['Collection']:
    """
    gets the collection by a location.

    :param loc: Location of the collection
    :return: collection at location or None
    """
    for c in COLLECTIONS:
        if c.loc_is_member(loc):
            return c
    return None


class CollectionScanThread(common.ProgressThread):
    """
    Scans the collection
    """

    def __init__(self, collection, startup_scan=False, force_update=False):
        """
        Initializes the thread

        :param collection: the collection to scan
        :param startup_scan: Only scan libraries scanned at startup
        :param force_update: Update files regardless whether they've changed
        """
        common.ProgressThread.__init__(self)

        self.startup_scan = startup_scan
        self.force_update = force_update
        self.collection = collection

    def stop(self):
        """
        Stops the thread
        """
        self.collection.stop_scan()
        common.ProgressThread.stop(self)

    def run(self):
        """
        Runs the thread
        """
        event.add_callback(self.on_scan_progress_update, 'scan_progress_update')

        self.collection.rescan_libraries(
            startup_only=self.startup_scan, force_update=self.force_update
        )

        event.remove_callback(self.on_scan_progress_update, 'scan_progress_update')

    def on_scan_progress_update(self, type, collection, progress):
        """
        Notifies about progress changes
        """
        if progress < 100:
            self.emit('progress-update', progress)
        else:
            self.emit('done')


class Collection(trax.TrackDB):
    """
    Manages a persistent track database.

    Simple usage:

    >>> from xl.collection import *
    >>> from xl.trax import search
    >>> collection = Collection("Test Collection")
    >>> collection.add_library(Library("./tests/data"))
    >>> collection.rescan_libraries()
    >>> tracks = [i.track for i in search.search_tracks_from_string(
    ...     collection, ('artist==TestArtist'))]
    >>> print(len(tracks))
    5
    """

    def __init__(self, name, location=None, pickle_attrs=[]):
        global COLLECTIONS
        self.libraries: Dict[str, Library] = {}
        self._scanning = False
        self._scan_stopped = False
        self._running_count = 0
        self._running_total_count = 0
        self._frozen = False
        self._libraries_dirty = False
        pickle_attrs += ['_serial_libraries']
        trax.TrackDB.__init__(self, name, location=location, pickle_attrs=pickle_attrs)
        COLLECTIONS.add(self)

    def freeze_libraries(self) -> None:
        """
        Prevents "libraries_modified" events from being sent from individual
        add and remove library calls.

        Call this before making bulk changes to the libraries. Call
        thaw_libraries when you are done; this sends a single event if the
        libraries were modified.
        """
        self._frozen = True

    def thaw_libraries(self) -> None:
        """
        Re-allow "libraries_modified" events from being sent from individual
        add and remove library calls. Also sends a "libraries_modified"
        event if the libraries have ben modified since the last call to
        freeze_libraries.
        """
        # TODO: This method should probably be synchronized.
        self._frozen = False
        if self._libraries_dirty:
            self._libraries_dirty = False
            event.log_event('libraries_modified', self, None)

    def add_library(self, library: 'Library') -> None:
        """
        Add this library to the collection

        :param library: the library to add
        """
        loc = library.get_location()
        if loc not in self.libraries:
            self.libraries[loc] = library
            library.set_collection(self)
        self.serialize_libraries()
        self._dirty = True

        if self._frozen:
            self._libraries_dirty = True
        else:
            event.log_event('libraries_modified', self, None)

    def remove_library(self, library: 'Library') -> None:
        """
        Remove a library from the collection

        :param library: the library to remove
        """
        for k, v in self.libraries.items():
            if v == library:
                del self.libraries[k]
                break

        to_rem = []
        if "://" not in library.location:
            location = "file://" + library.location
        else:
            location = library.location
        for tr in self.tracks:
            if tr.startswith(location):
                to_rem.append(self.tracks[tr]._track)
        self.remove_tracks(to_rem)

        self.serialize_libraries()
        self._dirty = True

        if self._frozen:
            self._libraries_dirty = True
        else:
            event.log_event('libraries_modified', self, None)

    def stop_scan(self):
        """
        Stops the library scan
        """
        self._scan_stopped = True

    def get_libraries(self) -> List['Library']:
        """
        Gets a list of all the Libraries associated with this
        Collection
        """
        return list(self.libraries.values())

    def rescan_libraries(self, startup_only=False, force_update=False):
        """
        Rescans all libraries associated with this Collection
        """
        if self._scanning:
            raise Exception("Collection is already being scanned")
        if len(self.libraries) == 0:
            event.log_event('scan_progress_update', self, 100)
            return  # no libraries, no need to scan :)

        self._scanning = True
        self._scan_stopped = False

        self.file_count = -1  # negative means we dont know it yet

        self.__count_files()

        scan_interval = 20

        for library in self.libraries.values():

            if (
                not force_update
                and startup_only
                and not (library.monitored and library.startup_scan)
            ):
                continue

            event.add_callback(self._progress_update, 'tracks_scanned', library)
            library.rescan(notify_interval=scan_interval, force_update=force_update)
            event.remove_callback(self._progress_update, 'tracks_scanned', library)
            self._running_total_count += self._running_count
            if self._scan_stopped:
                break
        else:  # didnt break
            try:
                if self.location is not None:
                    self.save_to_location()
            except AttributeError:
                logger.exception("Exception occurred while saving")

        event.log_event('scan_progress_update', self, 100)

        self._running_total_count = 0
        self._running_count = 0
        self._scanning = False
        self.file_count = -1

    @common.threaded
    def __count_files(self):
        file_count = 0
        for library in self.libraries.values():
            if self._scan_stopped:
                self._scanning = False
                return
            file_count += library._count_files()
        self.file_count = file_count
        logger.debug("File count: %s", self.file_count)

    def _progress_update(self, type, library, count):
        """
        Called when a progress update should be emitted while scanning
        tracks
        """
        self._running_count = count
        count = count + self._running_total_count

        if self.file_count < 0:
            event.log_event('scan_progress_update', self, 0)
            return

        try:
            event.log_event(
                'scan_progress_update',
                self,
                count / self.file_count * 100,
            )
        except ZeroDivisionError:
            pass

    def serialize_libraries(self):
        """
        Save information about libraries

        Called whenever the library's settings are changed
        """
        _serial_libraries = []
        for k, v in self.libraries.items():
            l = {}
            l['location'] = v.location
            l['monitored'] = v.monitored
            l['realtime'] = v.monitored
            l['scan_interval'] = v.scan_interval
            l['startup_scan'] = v.startup_scan
            _serial_libraries.append(l)
        return _serial_libraries

    def unserialize_libraries(self, _serial_libraries):
        """
        restores libraries from their serialized state.

        Should only be called once, from the constructor.
        """
        for l in _serial_libraries:
            self.add_library(
                Library(
                    l['location'],
                    l.get('monitored', l.get('realtime')),
                    l['scan_interval'],
                    l.get('startup_scan', True),
                )
            )

    _serial_libraries = property(serialize_libraries, unserialize_libraries)

    def close(self):
        """
        close the collection. does any work like saving to disk,
        closing network connections, etc.
        """
        # TODO: make close() part of trackdb
        COLLECTIONS.remove(self)

    def delete_tracks(self, tracks: Iterable[trax.Track]) -> None:
        for tr in tracks:
            for prefix, lib in self.libraries.items():
                lib.delete(tr.get_loc_for_io())


class LibraryMonitor(GObject.GObject):
    """
    Monitors library locations for changes
    """

    __gproperties__ = {
        'monitored': (
            GObject.TYPE_BOOLEAN,
            'monitoring state',
            'Whether to monitor this library',
            False,
            GObject.ParamFlags.READWRITE,
        )
    }
    __gsignals__ = {
        'location-added': (GObject.SignalFlags.RUN_LAST, None, [Gio.File]),
        'location-removed': (GObject.SignalFlags.RUN_LAST, None, [Gio.File]),
    }

    def __init__(self, library):
        """
        :param library: the library to monitor
        :type library: :class:`Library`
        """
        GObject.GObject.__init__(self)

        self.__library = library
        self.__root = Gio.File.new_for_uri(library.location)
        self.__monitored = False
        self.__monitors = {}
        self.__queue = {}
        self.__lock = threading.RLock()

    def do_get_property(self, property):
        """
        Gets GObject properties
        """
        if property.name == 'monitored':
            return self.__monitored
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, property, value):
        """
        Sets GObject properties
        """
        if property.name == 'monitored':
            if value != self.__monitored:
                self.__monitored = value
                update_thread = threading.Thread(target=self.__update_monitors)
                update_thread.daemon = True
                GLib.idle_add(update_thread.start)
        else:
            raise AttributeError('unknown property %s' % property.name)

    def __update_monitors(self):
        """
        Sets up or removes library monitors
        """
        with self.__lock:
            if self.props.monitored:
                logger.debug('Setting up library monitors')

                for directory in common.walk_directories(self.__root):
                    monitor = directory.monitor_directory(
                        Gio.FileMonitorFlags.NONE, None
                    )
                    monitor.connect('changed', self.on_location_changed)
                    self.__monitors[directory] = monitor

                    self.emit('location-added', directory)
            else:
                logger.debug('Removing library monitors')

                for directory, monitor in self.__monitors.items():
                    monitor.cancel()

                    self.emit('location-removed', directory)

                self.__monitors = {}

    def __process_change_queue(self, gfile):
        if gfile in self.__queue:
            added_tracks = trax.util.get_tracks_from_uri(gfile.get_uri())
            for tr in added_tracks:
                tr.read_tags()
            self.__library.collection.add_tracks(added_tracks)
            del self.__queue[gfile]

    def on_location_changed(self, monitor, gfile, other_gfile, event):
        """
        Updates the library on changes of the location
        """

        if event == Gio.FileMonitorEvent.CHANGES_DONE_HINT:
            self.__process_change_queue(gfile)
        elif (
            event == Gio.FileMonitorEvent.CREATED
            or event == Gio.FileMonitorEvent.CHANGED
        ):

            # Enqueue tracks retrieval
            if gfile not in self.__queue:
                self.__queue[gfile] = True

                # File monitor only emits the DONE_HINT when using inotify,
                # and only on single files. Give it some time, but don't
                # lose the change notification
                GLib.timeout_add(500, self.__process_change_queue, gfile)

            # Set up new monitor if directory
            fileinfo = gfile.query_info(
                'standard::type', Gio.FileQueryInfoFlags.NONE, None
            )

            if (
                fileinfo.get_file_type() == Gio.FileType.DIRECTORY
                and gfile not in self.__monitors
            ):

                for directory in common.walk_directories(gfile):
                    monitor = directory.monitor_directory(
                        Gio.FileMonitorFlags.NONE, None
                    )
                    monitor.connect('changed', self.on_location_changed)
                    self.__monitors[directory] = monitor

                    self.emit('location-added', directory)

        elif event == Gio.FileMonitorEvent.DELETED:
            removed_tracks = []

            track = trax.Track(gfile.get_uri())

            if track in self.__library.collection:
                # Deleted file was a regular track
                removed_tracks += [track]
            else:
                # Deleted file was most likely a directory
                for track in self.__library.collection:
                    track_gfile = Gio.File.new_for_uri(track.get_loc_for_io())

                    if track_gfile.has_prefix(gfile):
                        removed_tracks += [track]

            self.__library.collection.remove_tracks(removed_tracks)

            # Remove obsolete monitors
            removed_directories = [
                d for d in self.__monitors if d == gfile or d.has_prefix(gfile)
            ]

            for directory in removed_directories:
                self.__monitors[directory].cancel()
                del self.__monitors[directory]

                self.emit('location-removed', directory)


class Library:
    """
    Scans and watches a folder for tracks, and adds them to
    a Collection.

    Simple usage:

    >>> from xl.collection import *
    >>> c = Collection("TestCollection")
    >>> l = Library("./tests/data")
    >>> c.add_library(l)
    >>> l.rescan()
    True
    >>> print(c.get_libraries()[0].location)
    ./tests/data
    >>> print(len(list(c.search('artist="TestArtist"'))))
    5
    >>>
    """

    def __init__(
        self,
        location: str,
        monitored: bool = False,
        scan_interval: int = 0,
        startup_scan: bool = False,
    ):
        """
        Sets up the Library

        :param location: the directory this library will scan
        :param monitored: whether the library should update its
            collection at changes within the library's path
        :param scan_interval: the interval for automatic rescanning
        """
        self.location = location
        self.scan_interval = scan_interval
        self.scan_id = None
        self.scanning = False
        self._startup_scan = startup_scan
        self.monitor = LibraryMonitor(self)
        self.monitor.props.monitored = monitored

        self.collection: Optional[Collection] = None
        self.set_rescan_interval(scan_interval)

    def set_location(self, location: str) -> None:
        """
        Changes the location of this Library

        :param location: the new location to use
        """
        self.location = location

    def get_location(self) -> str:
        """
        Gets the current location associated with this Library

        :return: the current location
        """
        return self.location

    def set_collection(self, collection) -> None:
        self.collection = collection

    def get_monitored(self) -> bool:
        """
        Whether the library should be monitored for changes
        """
        return self.monitor.props.monitored

    def set_monitored(self, monitored: bool) -> None:
        """
        Enables or disables monitoring of the library

        :param monitored: Whether to monitor the library
        :type monitored: bool
        """
        self.monitor.props.monitored = monitored
        self.collection.serialize_libraries()
        self.collection._dirty = True

    monitored = property(get_monitored, set_monitored)

    def get_rescan_interval(self) -> int:
        """
        :return: the scan interval in seconds
        """
        return self.scan_interval

    def set_rescan_interval(self, interval: int) -> None:
        """
        Sets the scan interval in seconds.  If the interval is 0 seconds,
        the scan interval is stopped

        :param interval: scan interval in seconds
        """

        if self.scan_id:
            GLib.source_remove(self.scan_id)
            self.scan_id = None

        if interval:
            self.scan_id = GLib.timeout_add_seconds(interval, self.rescan)

        self.scan_interval = interval

    def get_startup_scan(self) -> bool:
        return self._startup_scan

    def set_startup_scan(self, value: bool) -> None:
        self._startup_scan = value
        self.collection.serialize_libraries()
        self.collection._dirty = True

    startup_scan = property(get_startup_scan, set_startup_scan)

    def _count_files(self) -> int:
        """
        Counts the number of files present in this directory
        """
        count = 0
        for file in common.walk(Gio.File.new_for_uri(self.location)):
            if self.collection:
                if self.collection._scan_stopped:
                    break
            count += 1

        return count

    def _check_compilation(
        self,
        ccheck: Dict[str, Dict[str, Deque[str]]],
        compilations: MutableSequence[Tuple[str, str]],
        tr: trax.Track,
    ) -> None:
        """
        This is the hacky way to test to see if a particular track is a
        part of a compilation.

        Basically, if there is more than one track in a directory that has
        the same album but different artist, we assume that it's part of a
        compilation.

        :param ccheck: dictionary for internal use
        :param compilations: if a compilation is found, it'll be appended
            to this list
        :param tr: the track to check
        """
        # check for compilations
        if not settings.get_option('collection/file_based_compilations', True):
            return

        def joiner(value):
            if isinstance(value, list):
                return "\0".join(value)
            else:
                return value

        try:
            basedir = joiner(tr.get_tag_raw('__basedir'))
            album = joiner(tr.get_tag_raw('album'))
            artist = joiner(tr.get_tag_raw('artist'))
        except Exception:
            logger.warning("Error while checking for compilation: %s", tr)
            return
        if not basedir or not album or not artist:
            return
        album = album.lower()
        artist = artist.lower()
        try:
            if basedir not in ccheck:
                ccheck[basedir] = {}

            if album not in ccheck[basedir]:
                ccheck[basedir][album] = deque()
        except TypeError:
            logger.exception("Error adding to compilation")
            return

        if ccheck[basedir][album] and artist not in ccheck[basedir][album]:
            if not (basedir, album) in compilations:
                compilations.append((basedir, album))
                logger.debug("Compilation %r detected in %r", album, basedir)

        ccheck[basedir][album].append(artist)

    def update_track(
        self, gloc: Gio.File, force_update: bool = False
    ) -> Optional[trax.Track]:
        """
        Rescan the track at a given location

        :param gloc: the location
        :type gloc: :class:`Gio.File`
        :param force_update: Force update of file (default only updates file
                             when mtime has changed)

        returns: the Track object, None if it could not be updated
        """
        uri = gloc.get_uri()
        if not uri:  # we get segfaults if this check is removed
            return None

        tr = self.collection.get_track_by_loc(uri)
        if tr:
            tr.read_tags(force=force_update)
        else:
            tr = trax.Track(uri)
            if tr._scan_valid:
                self.collection.add(tr)

            # Track already existed. This fixes trax.get_tracks_from_uri
            # on windows, unknown why fix isnt needed on linux.
            elif not tr._init:
                self.collection.add(tr)
        return tr

    def rescan(
        self, notify_interval: Optional[int] = None, force_update: bool = False
    ):  # TODO: What return type?
        """
        Rescan the associated folder and add the contained files
        to the Collection
        """
        # TODO: use gio's cancellable support

        if self.collection is None:
            return True

        if self.scanning:
            return

        logger.info("Scanning library: %s", self.location)
        self.scanning = True
        libloc = Gio.File.new_for_uri(self.location)

        count = 0
        dirtracks = deque()
        compilations = deque()
        ccheck = {}
        for fil in common.walk(libloc):
            count += 1
            type = fil.query_info(
                "standard::type", Gio.FileQueryInfoFlags.NONE, None
            ).get_file_type()
            if type == Gio.FileType.DIRECTORY:
                if dirtracks:
                    for tr in dirtracks:
                        self._check_compilation(ccheck, compilations, tr)
                    for (basedir, album) in compilations:
                        base = basedir.replace('"', '\\"')
                        alb = album.replace('"', '\\"')
                        items = [
                            tr
                            for tr in dirtracks
                            if tr.get_tag_raw('__basedir') == base and
                            # FIXME: this is ugly
                            alb in "".join(tr.get_tag_raw('album') or []).lower()
                        ]
                        for item in items:
                            item.set_tag_raw('__compilation', (basedir, album))
                dirtracks = deque()
                compilations = deque()
                ccheck = {}
            elif type == Gio.FileType.REGULAR:
                tr = self.update_track(fil, force_update=force_update)
                if not tr:
                    continue

                if dirtracks is not None:
                    dirtracks.append(tr)
                    # do this so that if we have, say, a 4000-song folder
                    # we dont get bogged down trying to keep track of them
                    # for compilation detection. Most albums have far fewer
                    # than 110 tracks anyway, so it is unlikely that this
                    # restriction will affect the heuristic's accuracy.
                    # 110 was chosen to accomodate "top 100"-style
                    # compilations.
                    if len(dirtracks) > 110:
                        logger.debug(
                            "Too many files, skipping "
                            "compilation detection heuristic for %s",
                            fil.get_uri(),
                        )
                        dirtracks = None

            if self.collection and self.collection._scan_stopped:
                self.scanning = False
                logger.info("Scan canceled")
                return

            # progress update
            if notify_interval is not None and count % notify_interval == 0:
                event.log_event('tracks_scanned', self, count)

        # final progress update
        if notify_interval is not None:
            event.log_event('tracks_scanned', self, count)

        removals = deque()
        for tr in self.collection.tracks.values():
            tr = tr._track
            loc = tr.get_loc_for_io()
            if not loc:
                continue
            gloc = Gio.File.new_for_uri(loc)
            try:
                if not gloc.has_prefix(libloc):
                    continue
            except UnicodeDecodeError:
                logger.exception("Error decoding file location")
                continue

            if not gloc.query_exists(None):
                removals.append(tr)

        for tr in removals:
            logger.debug("Removing %s", tr)
            self.collection.remove(tr)

        logger.info("Scan completed: %s", self.location)
        self.scanning = False

    def add(self, loc: str, move: bool = False) -> None:
        """
        Copies (or moves) a file into the library and adds it to the
        collection
        """
        oldgloc = Gio.File.new_for_uri(loc)

        newgloc = Gio.File.new_for_uri(self.location).resolve_relative_path(
            oldgloc.get_basename()
        )

        if move:
            oldgloc.move(newgloc)
        else:
            oldgloc.copy(newgloc)
        tr = trax.Track(newgloc.get_uri())
        if tr._scan_valid:
            self.collection.add(tr)

    def delete(self, loc: str) -> None:
        """
        Deletes a file from the disk

        .. warning::
           This permanently deletes the file from the hard disk.
        """
        tr = self.collection.get_track_by_loc(loc)
        if tr:
            self.collection.remove(tr)
            loc = tr.get_loc_for_io()
            file = Gio.File.new_for_uri(loc)
            if not file.delete():
                logger.warning("Could not delete file %s.", loc)


class TransferQueue:
    def __init__(self, library: Library):
        self.library = library
        self.queue: List[trax.Track] = []
        self.current_pos = -1
        self.transferring = False
        self._stop = False

    def enqueue(self, tracks: Iterable[trax.Track]) -> None:
        self.queue.extend(tracks)

    def dequeue(self, tracks: Iterable[trax.Track]) -> None:
        if self.transferring:
            # FIXME: use a proper exception, and make this only error on
            # tracks that have already been transferred
            raise Exception("Cannot remove tracks while transferring")

        for t in tracks:
            try:
                self.queue.remove(t)
            except ValueError:
                pass

    def transfer(self) -> None:
        """
        Tranfer the queued tracks to the library.

        This is NOT asynchronous
        """
        self.transferring = True
        self.current_pos += 1
        try:
            while self.current_pos < len(self.queue) and not self._stop:
                track = self.queue[self.current_pos]
                loc = track.get_loc_for_io()
                self.library.add(loc)

                # TODO: make this be based on filesize not count
                progress = self.current_pos * 100 / len(self.queue)
                event.log_event('track_transfer_progress', self, progress)

                self.current_pos += 1
        finally:
            self.queue = []
            self.transferring = False
            self.current_pos = -1
            self._stop = False
            event.log_event('track_transfer_progress', self, 100)

    def cancel(self) -> None:
        """
        Cancel the current transfer
        """
        # TODO: make this stop mid-file as well?
        self._stop = True



# Copyright (C) 2009-2010 Aren Olson, 2014 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

'''
    If you don't care about whether the UI/exaile have finished loading,
    you can create a plugin by implementing the following functions:

        def enable(exaile):
            print("Hello, world!")
            testlib.sucess()

        def disable(exaile):
            print("Goodbye. :(")

        def teardown(exaile):
            # optional function
            print("Unhello, World!")

    Exaile 3.4 introduced a new way to write plugins which will eliminate
    a lot of unnecessary boilerplate for plugin authors.

    New-style plugins must have a variable in the module that is
    called 'plugin_class', which is a Python object with the following
    interface:

    class Foo:

        def enable(self, exaile):
            pass

        def disable(self, exaile):
            pass

    The object can also have the following optional functions:

        def on_gui_loaded(self):
            - This will be called when the GUI is ready, or
              immediately if already done

        def on_exaile_loaded(self):
            - This will be called when exaile is done loading, or
              immediately if already done

        def teardown(self, exaile):
            - This will be called when exaile is unloading

    Note that the old style of writing plugins is not going away, so you
    can continue to write plugins that way if its simpler.
'''


from . import testlib


class HelloWorld:
    def enable(self, exaile):
        print("Hello, world!")
        testlib.sucess()

    def disable(self, exaile):
        print("Goodbye. :(")

    def teardown(self, exaile):
        '''Optional function'''
        print("Unhello, World!")

    def on_exaile_loaded(self):
        '''Optional function'''
        print('Exaile loaded!')


plugin_class = HelloWorld



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gtk
from xl import providers


class Accelerator:
    __slots__ = ['name', 'keys', 'helptext', 'callback', 'key', 'mods']

    def __init__(self, keys, helptext, callback):
        self.name = keys  # only here because providers needs it
        self.keys = keys
        self.helptext = helptext
        self.callback = callback
        self.key, self.mods = Gtk.accelerator_parse(keys)


class AcceleratorManager(providers.ProviderHandler):
    def __init__(self, providername, accelgroup):
        self.accelgroup = accelgroup
        self.accelerators = {}
        providers.ProviderHandler.__init__(self, providername, simple_init=True)

    def on_provider_added(self, provider):
        self.accelgroup.connect(
            provider.key, provider.mods, Gtk.AccelFlags.VISIBLE, provider.callback
        )

        # Add accelerator to internal list, so we can enable and disable
        # them via enable_accelerators() and disable_accelerators()
        self.accelerators[(provider.key, provider.mods)] = provider

    def on_provider_removed(self, provider):
        self.accelgroup.disconnect_key(provider.key, provider.mods)

        # Remove accelerator from our internal list
        del self.accelerators[(provider.key, provider.mods)]

    ## Global accelerator enable/disable
    def disable_accelerators(self):
        for provider in self.accelerators.values():
            self.accelgroup.disconnect_key(provider.key, provider.mods)

    def enable_accelerators(self):
        for provider in self.accelerators.values():
            self.accelgroup.connect(
                provider.key, provider.mods, Gtk.AccelFlags.VISIBLE, provider.callback
            )



#
# Format is pretty simple. Provide the presets so that users
# can play around with 'proven' interesting visualizations
# without too much effort.
#
# Each preset is a tuple of (name, template_name, [tag_data])
# tag_data is currently a list of tag names
#

# TODO: add more interesting presets!

DEFAULT_PRESETS = [
    ('BPM Graph', 'force.tmpl.html', ['bpm']),
    ('BPM bar chart', 'bar_chart.tmpl.html', ['bpm']),
    ('GroupTagger bar chart', 'bar_chart.tmpl.html', ['__grouptagger']),
    ('Concept map Title/Genre', 'concept_map.tmpl.html', ['title', 'genre']),
    ('Concept map BPM/GroupTagger', 'concept_map.tmpl.html', ['bpm', '__grouptagger']),
]



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from xl.metadata._base import BaseFormat, CoverImage
from mutagen import mp4


class MP4Format(BaseFormat):
    MutagenType = mp4.MP4
    tag_mapping = {
        # fmt: off
        'title':       '\xa9nam',
        'artist':      '\xa9ART',
        'albumartist': '\x61ART',
        'album':       '\xa9alb',
        'composer':    '\xa9wrt',
        'genre':       '\xa9gen',
        'lyrics':      '\xa9lyr',
        'encodedby':   '\xa9too',
        'date':        '\xa9day',
        'tracknumber': 'trkn',
        'discnumber':  'disk',
        'copyright':   'cprt',
        'bpm':         'tmpo',
        'grouping':    '\xa9grp',
        'comment':     '\xa9cmt',
        'originaldate': '----:com.apple.iTunes:ORIGYEAR',
        'cover':       'covr',
        'language':    '----:com.apple.iTunes:LANGUAGE',
        # fmt: on
    }
    others = False
    writable = True

    def _get_tag(self, f, name):
        if name not in f:
            return []
        elif name == 'covr':
            ret = []
            for value in f[name]:
                if value.imageformat == mp4.MP4Cover.FORMAT_PNG:
                    mime = 'image/png'
                else:
                    mime = 'image/jpeg'
                ret.append(CoverImage(type=None, desc=None, mime=mime, data=value))
            return ret
        elif name in ['trkn', 'disk']:
            ret = []
            for value in f[name]:
                ret.append("%d/%d" % (value[0], value[1]))
            return ret
        else:
            return [t for t in f[name]]

    def _set_tag(self, f, name, value):
        if not isinstance(value, list):
            value = [value]
        if name in ['trkn', 'disk']:
            try:
                f[name] = []
                for val in value:
                    tmp = map(int, val.split('/'))
                    f[name].append(tuple(tmp))
            except (TypeError, ValueError):
                pass
        elif name == 'covr':
            f[name] = []

            for val in value:
                if val.mime == 'image/jpeg':
                    f[name].append(mp4.MP4Cover(val.data, mp4.MP4Cover.FORMAT_JPEG))
                elif val.mime == 'image/png':
                    f[name].append(mp4.MP4Cover(val.data, mp4.MP4Cover.FORMAT_JPEG))
                else:
                    raise ValueError(
                        'MP4 does not support cover image type %s' % val.type
                    )
        elif name == 'tmpo':
            f[name] = [int(v) for v in value]
        elif name == '----:com.apple.iTunes:ORIGYEAR':
            f[name] = [str(v) for v in value]
        else:
            f[name] = value


# vim: et sts=4 sw=4



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

from xl.nls import gettext as _
from xl import providers
from xlgui import guiutil
from xlgui.widgets import menu

# Custom tab style; fixes some Adwaita ugliness
TAB_CSS = Gtk.CssProvider()
TAB_CSS.load_from_data(
    b'''
    /* Most themes don't handle vertical notebooks well,
        so we override everything. */
    notebook.vertical tab {
        padding: 3px 4px 3px 4px;
    }
    notebook.vertical tab * {
        margin: 0;
        padding: 0;
    }
    notebook.vertical tab label {
        margin: 0 0 3px 0;
    }
    /* Remove gap between tabs */
    header.top tab, header.bottom tab {
        margin-left: -1px;
        margin-right: -1px;
    }
    header.left tab, header.right tab {
        margin-top: -1px;
        margin-bottom: -1px;
    }
    '''
)


def apply_css(widget):
    sc = widget.get_style_context()
    sc.add_provider(TAB_CSS, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    return sc


class SmartNotebook(Gtk.Notebook):
    def __init__(self, vertical=False):
        Gtk.Notebook.__init__(self)
        self.set_scrollable(True)
        self.connect('button-press-event', self.on_button_press)
        self.connect('popup-menu', self.on_popup_menu)
        self.connect('notify::tab-pos', self.__on_notify_tab_pos)
        self._add_tab_on_empty = True

        sc = apply_css(self)
        if vertical:
            sc.add_class('vertical')
            self.set_tab_pos(Gtk.PositionType.LEFT)

    def get_current_tab(self):
        current_page = self.get_current_page()
        if current_page == -1:
            return None
        return self.get_nth_page(current_page)

    def add_tab(self, tab, page, position=-1, switch=True):
        """
        Add a tab to the notebook. It will be given focus.

        :param tab: The tab to use
        :type tab: NotebookTab
        :param page: The page to use
        :type page: NotebookPage
        :param position: index to insert page at, or -1 for append
        :type position: int
        """
        self.insert_page(page, tab, position=position)
        tab.notebook = self
        self.set_tab_reorderable(page, page.reorderable)
        tab_pos = self.get_tab_pos()
        if tab_pos in (Gtk.PositionType.TOP, Gtk.PositionType.BOTTOM):
            self.child_set_property(page, 'tab-expand', True)
        tab.adjust_label_width(tab_pos)
        if switch:
            self.set_current_page(self.page_num(page))

    def add_default_tab(self):
        """
        Action taken when a generic "new tab" option is triggered.
        Subclasses need to override this if they want new tab
        functionality to work automatically.

        :return: The NotebookTab created, or None
        """
        pass

    def remove_page(self, page_num):
        """
        Overrides Gtk.Notebook.remove_page
        """
        if page_num == -1:
            page_num = self.get_n_pages() - 1
        tab = self.get_tab_label(self.get_nth_page(page_num))

        Gtk.Notebook.remove_page(self, page_num)
        tab.notebook = None

        if self._add_tab_on_empty and self.get_n_pages() == 0:
            self.add_default_tab()

    def remove_tab(self, tab):
        """
        Remove a specific NotebookTab from the notebook
        """
        page_num = self.page_num(tab.page)
        if page_num >= 0:
            self.remove_page(page_num)

    def set_add_tab_on_empty(self, add_tab_on_empty):
        """
        If set True, the SmartNotebook will always maintain at
        least one tab in the notebook
        """
        self._add_tab_on_empty = add_tab_on_empty

    def on_button_press(self, widget, event):
        if (
            event.type == Gdk.EventType.BUTTON_PRESS
            and event.button == Gdk.BUTTON_MIDDLE
        ):
            self.add_default_tab()

    def on_popup_menu(self, widget):
        page = self.get_current_tab()
        tab_label = self.get_tab_label(self.get_current_tab())
        page.tab_menu.popup(None, None, guiutil.position_menu, tab_label, 0, 0)
        return True

    def __on_notify_tab_pos(self, _widget, _param):
        tab_pos = self.get_tab_pos()
        expand = tab_pos in (Gtk.PositionType.TOP, Gtk.PositionType.BOTTOM)
        for i in range(self.get_n_pages()):
            page = self.get_nth_page(i)
            self.child_set_property(page, 'tab-expand', expand)
            tab = self.get_tab_label(page)  # type: NotebookTab
            tab.adjust_label_width(tab_pos)


class NotebookTab(Gtk.EventBox):
    """
    Class to represent a generic tab in a Gtk.Notebook.
    """

    reorderable = True

    def __init__(self, notebook, page, vertical=False):
        """
        :param notebook: The notebook this tab will belong to
        :type notebook: SmartNotebook
        :param page: The page this tab will be associated with
        :type page: NotebookPage
        :param vertical: Whether the tab contents are to be laid out vertically
        :type vertical: bool
        """
        Gtk.EventBox.__init__(self)
        self.set_visible_window(False)

        self.closable = True

        self.notebook = notebook
        self.page = page
        self.page.tab_menu.attach_to_widget(self.page, None)

        self.connect('button-press-event', self.on_button_press)

        self.vertical = vertical
        if vertical:
            box = Gtk.Box(spacing=2, orientation=Gtk.Orientation.VERTICAL)
        else:
            box = Gtk.Box(spacing=2)
        self.add(box)
        apply_css(box)

        self.icon = Gtk.Image()
        self.icon.set_no_show_all(True)
        apply_css(self.icon)

        self.label = Gtk.Label(label=self.page.get_page_name())
        self.label.set_tooltip_text(self.page.get_page_name())
        apply_css(self.label)

        if vertical:
            self.label.set_angle(90)
            self.label.props.valign = Gtk.Align.CENTER
            # Don't ellipsize but give a sane maximum length.
            self.label.set_max_width_chars(20)
        else:
            self.label.props.halign = Gtk.Align.CENTER
            self.label.set_ellipsize(Pango.EllipsizeMode.END)
            self.adjust_label_width(Gtk.PositionType.TOP)

        if self.can_rename():
            self.entry = entry = Gtk.Entry()
            entry.set_width_chars(self.label.get_max_width_chars())
            entry.set_text(self.label.get_text())
            border = Gtk.Border.new()
            border.left = 1
            border.right = 1
            entry.connect('activate', self.on_entry_activate)
            entry.connect('focus-out-event', self.on_entry_focus_out_event)
            entry.connect('key-press-event', self.on_entry_key_press_event)
            entry.set_no_show_all(True)
            apply_css(entry)

        self.button = button = Gtk.Button()
        button.set_relief(Gtk.ReliefStyle.NONE)
        button.set_halign(Gtk.Align.CENTER)
        button.set_valign(Gtk.Align.CENTER)
        button.set_focus_on_click(False)
        button.set_tooltip_text(_("Close Tab"))
        button.add(Gtk.Image.new_from_icon_name('window-close', Gtk.IconSize.MENU))
        button.connect('clicked', self.close)
        button.connect('button-press-event', self.on_button_press)
        apply_css(button)

        # pack the widgets in
        if vertical:
            box.pack_start(button, False, False, 0)
            box.pack_end(self.icon, False, False, 0)
            box.pack_end(self.label, True, True, 0)
            if self.can_rename():
                box.pack_end(self.entry, True, True, 0)

        else:
            box.pack_start(self.icon, False, False, 0)
            box.pack_start(self.label, True, True, 0)
            if self.can_rename():
                box.pack_start(self.entry, True, True, 0)
            box.pack_end(button, False, False, 0)

        page.set_tab(self)
        page.connect('name-changed', self.on_name_changed)

        box.show_all()

    def adjust_label_width(self, tab_pos):
        """Change the label's minimum width according to tab position"""
        if self.vertical:
            # Vertical tabs don't care about tab position.
            return
        if tab_pos in (Gtk.PositionType.TOP, Gtk.PositionType.BOTTOM):
            # The number of characters here seems to be approximate.
            # 4 is enough for around 2 characters and an ellipsis.
            self.label.set_width_chars(4)
        else:
            self.label.set_width_chars(20)

    def set_icon(self, pixbuf):
        """
        Set the primary icon for the tab.

        :param pixbuf: The icon to use, or None to hide
        :type pixbuf: :class:`GdkPixbuf.Pixbuf`
        """
        if pixbuf is None:
            self.icon.set_property("visible", False)
        else:
            self.icon.set_from_pixbuf(pixbuf)
            self.icon.set_property("visible", True)

    def set_closable(self, closable):
        self.closable = closable
        self.button.set_sensitive(closable)

    def on_button_press(self, widget, event):
        """
        Handles mouse button events on the tab.

        Typically triggers renaming, closing and menu.
        """
        if (
            event.button == Gdk.BUTTON_PRIMARY
            and event.type == Gdk.EventType._2BUTTON_PRESS
        ):
            self.start_rename()
        elif event.button == Gdk.BUTTON_MIDDLE:
            self.close()
        elif event.triggers_context_menu():
            self.page.tab_menu.popup(None, None, None, None, event.button, event.time)
            return True

    def on_entry_activate(self, entry):
        """
        Handles end of editing and triggers the actual rename.
        """
        self.entry.props.editing_canceled = False
        self.end_rename()

    def on_entry_focus_out_event(self, widget, event):
        """
        Make defocusing the rename box equivalent to activating it.
        """
        if not self.entry.props.editing_canceled:
            widget.activate()

    def on_entry_key_press_event(self, widget, event):
        """
        Cancel rename if Escape is pressed
        """
        if event.keyval == Gdk.KEY_Escape:
            self.entry.props.editing_canceled = True
            self.end_rename()
            return True

    def on_name_changed(self, *args):
        self.label.set_text(self.page.get_page_name())

    def start_rename(self):
        """
        Initiates the renaming of a tab, if the page supports this.
        """
        if not self.can_rename():
            return
        self.entry.set_text(self.page.get_page_name())
        self.label.hide()
        self.button.hide()
        self.entry.show()
        self.entry.select_region(0, -1)
        self.entry.grab_focus()

    def end_rename(self, cancel=False):
        """
        Finishes or cancels the renaming
        """
        name = self.entry.get_text()
        if name.strip() != "" and not self.entry.props.editing_canceled:
            self.page.set_page_name(name)
            self.label.set_text(name)
            self.label.set_tooltip_text(name)

        self.entry.hide()
        self.label.show()
        self.button.show()

        self.entry.props.editing_canceled = False

    def can_rename(self):
        return hasattr(self.page, 'set_page_name')

    def close(self, *args):
        if self.closable and not self.page.emit('closing'):
            self.notebook.remove_page(self.notebook.page_num(self.page))


class NotebookPage(Gtk.Box):
    """
    Base class representing a page. Should never be used directly.
    """

    menu_provider_name = 'tab-context'  # override this in subclasses
    reorderable = True
    __gsignals__ = {
        'name-changed': (GObject.SignalFlags.RUN_LAST, None, ()),
        'closing': (GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN, ()),
    }

    def __init__(self, child=None, page_name=None, menu_provider_name=None):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        # sometimes you just want to create a page
        if menu_provider_name is not None:
            self.menu_provider_name = menu_provider_name

        self.tab = None
        self.tab_menu = menu.ProviderMenu(self.menu_provider_name, self)

        if child is not None:
            self.pack_start(child, True, True, 0)

        if page_name is not None:
            self.page_name = page_name

    def focus(self):
        """
        Grabs focus for this page. Should be overriden in subclasses.
        """
        self.grab_focus()

    def get_page_name(self):
        """
        Returns the name of this tab. Should be overriden in subclasses.

        Subclasses can also implement set_page_name(self, name) to allow
        renaming, but this is not mandatory.
        """
        if hasattr(self, 'page_name'):
            return self.page_name
        return "UNNAMED PAGE"

    def set_tab(self, tab):
        """
        Set the tab that holds this page.  This will be called directly
        from the tab itself when it is created, and should not be used
        outside of that.
        """
        self.tab = tab

    def is_current_page(self):
        """
        Returns True if this page is the currently-visible page in
        the Notebook.
        """
        return self.tab.get_nth_page(self.tab.get_current_page()) == self

    def name_changed(self):
        self.emit('name-changed')


class NotebookAction:
    """
    A custom action to be placed to the left or right of tabs in a notebook
    """

    name = None
    position = Gtk.PackType.END

    def __init__(self, notebook):
        self.notebook = notebook


class NotebookActionService(providers.ProviderHandler):
    """
    Provides interface for action widgets to be dynamically attached
    detached from notebooks.

    Actions are widgets placed to the left or right of tabs on a notebook.
    """

    def __init__(self, notebook, servicename):
        """
        :param notebook:     Notebook to attach to
        :param servicename:  Provider service name to use
        """

        providers.ProviderHandler.__init__(self, servicename, notebook)

        self.notebook = notebook

        # Try to set up action widgets
        lw = Gtk.Box(spacing=3)
        lw.pack_end(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), False, False, 0)
        rw = Gtk.Box(spacing=3)
        lw.pack_start(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), False, False, 0)

        notebook.set_action_widget(lw, Gtk.PackType.START)
        notebook.set_action_widget(rw, Gtk.PackType.END)

        self.__actions = {}
        for provider in self.get_providers():
            self.on_provider_added(provider)

    def on_provider_added(self, provider):
        """
        Adds actions on provider addition
        """
        try:
            actions_box = self.notebook.get_action_widget(provider.position)
        except AttributeError:
            pass
        else:
            self.__actions[provider.name] = provider(self.notebook)
            actions_box.pack_start(self.__actions[provider.name], False, False, 0)
            actions_box.show_all()

    def on_provider_removed(self, provider):
        """
        Removes actions on provider removal
        """
        try:
            actions_box = self.notebook.get_action_widget(provider.position)
        except AttributeError:
            pass
        else:
            action = self.__actions[provider.name]
            actions_box.remove(action)
            action.destroy()

            del self.__actions[provider.name]



# Copyright (C) 2009-2010 Abhishek Mukherjee <abhishek.mukher.g@gmail.com>
# Copyright (C) 2017 Christian Stadelmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _('Notify')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "notifyprefs_pane.ui")


class ResizeCovers(widgets.CheckPreference, widgets.CheckConditional):
    default = False
    name = 'plugin/notify/resize'
    condition_preference_name = 'plugin/notify/show_covers'


class ShowCovers(widgets.CheckPreference):
    default = True
    name = 'plugin/notify/show_covers'
    pre_migration_name = 'plugin/notifyosd/covers'


class NotifyPause(widgets.CheckPreference):
    default = True
    name = 'plugin/notify/notify_pause'
    pre_migration_name = 'plugin/notifyosd/notify_pause'


class UseMediaIcons(widgets.CheckPreference):
    default = True
    name = 'plugin/notify/use_media_icons'
    pre_migration_name = 'plugin/notifyosd/media_icons'


class TrayHover(widgets.CheckPreference, widgets.CheckConditional):
    default = False
    name = 'plugin/notify/tray_hover'
    pre_migration_name = 'plugin/notifyosd/tray_hover'
    condition_preference_name = 'gui/use_tray'


class ShowWhenFocused(widgets.CheckPreference):
    default = False
    name = 'plugin/notify/show_when_focused'
    pre_migration_name = 'plugin/notifyosd/show_when_focused'



# Copyright (C) 2014 Dustin Spicuzza <dustin@virtualroadside.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

#
# Contains definitions for all of the context menus for panels
#
# Each panel pulls in menu items from different providers, so that
# plugins can easily hook all of the panels for track operations
# without a lot of work.
#
# To hook into a panel, just call 'register' on the menu item with
# the appropriate provider key.
#
# Each panel context has the following keys available:
#
#    - selection-empty
#       Use this for conditions to determine whether to show the menu
#       item or not.
#
#    - selected-tracks
#       Will return all selected tracks, even if a playlist is selected
#
# Other context keys are available for specific panels
#

from xl.nls import gettext as _
from xl import common
from xlgui.widgets import menu, menuitems

### Generic track selection menus


def __create_track_panel_menus():

    items = []

    items.append(menuitems.EnqueueMenuItem('enqueue', after=['top-sep']))

    items.append(menuitems.AppendMenuItem('append', after=[items[-1].name]))
    items.append(menuitems.ReplaceCurrentMenuItem('replace', after=[items[-1].name]))
    items.append(menuitems.RatingMenuItem('rating', after=[items[-1].name]))

    items.append(menu.simple_separator('tp-sep', after=[items[-1].name]))

    items.append(menuitems.PropertiesMenuItem('properties', after=[items[-1].name]))

    for item in items:
        item.register('track-panel-menu')


__create_track_panel_menus()


class TrackPanelMenu(menu.ProviderMenu):
    """
    Context menu when a track is clicked on a panel

    Provider key: track-panel-menu
    """

    def __init__(self, parent):
        menu.ProviderMenu.__init__(self, 'track-panel-menu', parent)

    def get_context(self):
        context = common.LazyDict(self._parent)
        context[
            'selected-tracks'
        ] = lambda name, parent: parent.tree.get_selected_tracks()
        context[
            'selection-empty'
        ] = lambda name, parent: parent.tree.get_selection_empty()
        return context


### Collection panel menus


def __create_collection_panel_context_menu():
    def collection_delete_tracks_func(panel, context, tracks):
        panel.collection.delete_tracks(tracks)

    items = []
    items.append(menu.simple_separator('cp-sep', after=['properties']))
    items.append(
        menuitems.OpenDirectoryMenuItem('open-directory', after=[items[-1].name])
    )
    items.append(
        menuitems.TrashMenuItem(
            'trash-tracks',
            after=[items[-1].name],
            delete_tracks_func=collection_delete_tracks_func,
        )
    )

    for item in items:
        item.register('collection-panel-context-menu')


__create_collection_panel_context_menu()


class CollectionContextMenu(menu.MultiProviderMenu):
    """
    Context menu when a collection track is clicked

    Provider keys: track-panel-menu, collection-panel-context-menu
    """

    def __init__(self, panel):
        menu.MultiProviderMenu.__init__(
            self, ['track-panel-menu', 'collection-panel-context-menu'], panel
        )

    def get_context(self):
        context = common.LazyDict(self._parent)
        context[
            'selected-tracks'
        ] = lambda name, parent: parent.tree.get_selected_tracks()
        context[
            'selection-empty'
        ] = lambda name, parent: parent.tree.get_selection_empty()
        return context


### Files panel menu


def __create_files_panel_context_menu():
    def trash_tracks_func(parent, context, tracks):
        menuitems.generic_trash_tracks_func(parent, context, tracks)
        parent.refresh(None)

    items = []
    items.append(menu.simple_separator('fp-sep', after=['properties']))
    items.append(
        menuitems.OpenDirectoryMenuItem('open-directory', after=[items[-1].name])
    )
    items.append(
        menuitems.TrashMenuItem(
            'trash-tracks', after=[items[-1].name], trash_tracks_func=trash_tracks_func
        )
    )

    for item in items:
        item.register('files-panel-context-menu')


__create_files_panel_context_menu()


class FilesContextMenu(menu.MultiProviderMenu):
    """
    Context menu when a files panel track is clicked

    Provider keys: track-panel-menu, files-panel-context-menu
    """

    def __init__(self, panel):
        menu.MultiProviderMenu.__init__(
            self, ['track-panel-menu', 'files-panel-context-menu'], panel
        )

    def get_context(self):
        context = common.LazyDict(self._parent)
        context[
            'needs-computing'
        ] = lambda name, parent: parent.tree.get_selection_is_computed()
        context[
            'selected-tracks'
        ] = lambda name, parent: parent.tree.get_selected_tracks()
        context[
            'selection-empty'
        ] = lambda name, parent: parent.tree.get_selection_empty()

        return context


### Playlist panel menus


def __create_playlist_panel_menus():

    # w, n, o, c: window, name, parent, context

    menu.simple_menu_item(
        'new-playlist',
        [],
        _('_New Playlist'),
        'tab-new',
        lambda w, n, o, c: o.add_new_playlist(),
    ).register('playlist-panel-menu')

    menu.simple_menu_item(
        'new-smart-playlist',
        ['new-playlist'],
        _('New _Smart Playlist'),
        'tab-new',
        lambda w, n, o, c: o.add_smart_playlist(),
    ).register('playlist-panel-menu')

    menu.simple_menu_item(
        'import-playlist',
        ['new-smart-playlist'],
        _('_Import Playlist'),
        'document-open',
        lambda w, n, o, c: o.import_playlist(),
    ).register('playlist-panel-menu')

    menu.simple_separator('top-sep', after=['import-playlist']).register(
        'playlist-panel-menu'
    )


__create_playlist_panel_menus()


class PlaylistPanelMenu(menu.ProviderMenu):
    """
    Menu for xlgui.panel.playlists.PlaylistsPanel, for when the
    user does not click on a playlist/track.  The default menu.

    Provider key: playlist-panel-menu
    """

    def __init__(self, parent):
        menu.ProviderMenu.__init__(self, 'playlist-panel-menu', parent)


def __create_playlist_panel_playlist_menus():

    items = []

    items.append(menu.simple_separator('pp-top-sep', ['properties']))

    items.append(
        menuitems.RenamePlaylistMenuItem('rename-playlist', after=[items[-1].name])
    )
    items.append(
        menuitems.EditPlaylistMenuItem('edit-playlist', after=[items[-1].name])
    )
    items.append(
        menuitems.ExportPlaylistMenuItem('export-playlist', after=[items[-1].name])
    )
    items.append(
        menuitems.ExportPlaylistFilesMenuItem('export-files', after=[items[-1].name])
    )

    items.append(menu.simple_separator('pp-sep', after=[items[-1].name]))

    items.append(
        menuitems.DeletePlaylistMenuItem('delete-playlist', after=[items[-1].name])
    )

    for item in items:
        item.register('playlist-panel-context-menu')


__create_playlist_panel_playlist_menus()


class PlaylistsPanelPlaylistMenu(menu.MultiProviderMenu):
    """
    Context menu when a playlist is clicked

    Provider keys: playlist-panel-menu, track-panel-menu, playlist-panel-context-menu
    """

    def __init__(self, parent):
        menu.MultiProviderMenu.__init__(
            self,
            ['playlist-panel-menu', 'track-panel-menu', 'playlist-panel-context-menu'],
            parent,
        )

    def get_context(self):
        context = common.LazyDict(self._parent)
        context[
            'needs-computing'
        ] = lambda name, parent: parent.tree.get_selection_is_computed()
        context[
            'selected-playlist'
        ] = lambda name, parent: parent.tree.get_selected_page(raw=True)
        context[
            'selected-tracks'
        ] = lambda name, parent: parent.tree.get_selected_tracks()
        context[
            'selection-empty'
        ] = lambda name, parent: parent.tree.get_selection_empty()
        return context


### Radio panel menu


def __create_radio_panel_menus():

    # w, n, o, c: window, name, parent, context

    menu.simple_menu_item(
        'new-station',
        [],
        _('_New Station'),
        'list-add',
        lambda w, n, o, c: o._on_add_button_clicked(),
    ).register('radio-panel-menu')


__create_radio_panel_menus()


class RadioPanelPlaylistMenu(menu.MultiProviderMenu):
    """
    Context menu when a playlist is clicked
    """

    def __init__(self, parent):
        menu.MultiProviderMenu.__init__(
            self,
            ['radio-panel-menu', 'track-panel-menu', 'playlist-panel-context-menu'],
            parent,
        )

    def get_context(self):
        context = common.LazyDict(self._parent)
        context[
            'selected-playlist'
        ] = lambda name, parent: parent.tree.get_selected_page(raw=True)
        context[
            'selected-tracks'
        ] = lambda name, parent: parent.tree.get_selected_tracks()
        context[
            'selection-empty'
        ] = lambda name, parent: parent.tree.get_selection_empty()
        return context



# winmmkeys - Adds support for multimedia keys in Win32.
# Copyright (C) 2007, 2010, 2016  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


class WinmmkeysPlugin:
    def enable(self, exaile):
        self.exaile = exaile

    def on_gui_loaded(self):
        try:
            self.handler = HotkeyHandler_PyHook(self.exaile)
        except ImportError:
            self.handler = HotkeyHandler_Keyboard(self.exaile)

    def disable(self, exaile):
        if hasattr(self, 'handler'):
            self.handler.disable()
            del self.handler
        del self.exaile


plugin_class = WinmmkeysPlugin


class HotkeyHandler_PyHook:
    def __init__(self, exaile):
        import pyHook
        from xl.player import PLAYER, QUEUE

        key_map = {
            'Launch_Media_Select': exaile.gui.main.window.present,
            'Media_Stop': PLAYER.stop,
            'Media_Play_Pause': PLAYER.toggle_pause,
            'Media_Next_Track': QUEUE.next,
            'Media_Prev_Track': QUEUE.prev,
        }

        def on_key_down(event):
            # NOTE: Because we capture every single key in the system, the
            # following test will fail almost 100% of the time. That's why we
            # use a very simple test that fails fast rather than a try-except
            # block.
            if event.Key in key_map:
                key_map[event.Key]()
                return False  # Indicate that we've handled the key.
            return True

        self.hook_manager = man = pyHook.HookManager()
        man.KeyDown = on_key_down
        man.HookKeyboard()

    def __del__(self):
        self.disable()

    def disable(self):
        if hasattr(self, 'hook_manager'):
            self.hook_manager.UnhookKeyboard()
            del self.hook_manager


class HotkeyHandler_Keyboard:
    def __init__(self, exaile):
        import keyboard
        from xl.player import PLAYER, QUEUE

        self.handlers = [
            # use lambda here because gi function isn't hashable
            keyboard.add_hotkey(
                'select media', lambda: exaile.gui.main.window.present()
            ),
            keyboard.add_hotkey('stop media', PLAYER.stop),
            keyboard.add_hotkey('play/pause media', PLAYER.toggle_pause),
            keyboard.add_hotkey('next track', QUEUE.next),
            keyboard.add_hotkey('previous track', QUEUE.prev),
        ]

    def __del__(self):
        self.disable()

    def disable(self):
        if hasattr(self, 'handlers'):
            import keyboard

            for handler in self.handlers:
                keyboard.remove_hotkey(handler)
            del self.handlers


# vi: et sts=4 sw=4 tw=80



# mpris2 - Support MPRIS 2 in Exaile
# Copyright (C) 2015-2016  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from gi.repository import Gio

from . import dbushelper
from . import mprisobject


# MPRIS 2.2 partial introspection data.
# This contains some Exaile-specific optimizations, e.g. Rate is marked as
# const because Exaile cannot change playback rate.
# The order of interfaces is important; later we refer to them by index.
MPRIS_INTROSPECTION = '''\
<node>
  <interface name='org.mpris.MediaPlayer2'>
    <method name='Raise' />
    <method name='Quit'>
      <annotation name='org.freedesktop.DBus.Method.NoReply' value='true' />
    </method>
    <property name='CanQuit' type='b' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='FullScreen' type='b' access='readwrite' />
    <property name='CanSetFullScreen' type='b' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='CanRaise' type='b' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='HasTrackList' type='b' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='Identity' type='s' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='DesktopEntry' type='s' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='SupportedUriSchemes' type='as' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='SupportedMimeTypes' type='as' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
  </interface>
  <interface name='org.mpris.MediaPlayer2.Player'>
    <method name='Next' />
    <method name='Previous' />
    <method name='Pause' />
    <method name='PlayPause' />
    <method name='Stop' />
    <method name='Play' />
    <method name='Seek'>
      <arg name='Offset' type='x' direction='in' />
    </method>
    <method name='SetPosition'>
      <arg name='TrackId' type='o' direction='in' />
      <arg name='Position' type='x' direction='in' />
    </method>
    <method name='OpenUri'>
      <arg name='Uri' type='s' direction='in' />
    </method>
    <signal name='Seeked'>
      <arg name='Position' type='x' />
    </signal>
    <property name='PlaybackStatus' type='s' access='read' />
    <property name='LoopStatus' type='s' access='readwrite' />
    <property name='Rate' type='d' access='readwrite'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='Shuffle' type='b' access='readwrite' />
    <property name='Metadata' type='a{sv}' access='read' />
    <property name='Volume' type='d' access='readwrite' />
    <property name='Position' type='x' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='false' />
    </property>
    <property name='MinimumRate' type='d' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='MaximumRate' type='d' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
    <property name='CanGoNext' type='b' access='read' />
    <property name='CanGoPrevious' type='b' access='read' />
    <property name='CanPlay' type='b' access='read' />
    <property name='CanPause' type='b' access='read' />
    <property name='CanSeek' type='b' access='read' />
    <property name='CanControl' type='b' access='read'>
      <annotation name='org.freedesktop.DBus.Property.EmitsChangedSignal' value='const' />
    </property>
  </interface>
</node>
'''


class MprisPlugin:
    def enable(self, exaile):
        self.handler = MprisHandler(exaile)

    def disable(self, exaile):
        self.handler.disconnect()
        del self.handler

    def teardown(self, exaile):
        self.handler.teardown()


plugin_class = MprisPlugin


class MprisHandler:
    root_interface = player_interface = None

    def __init__(self, exaile):
        if not self.root_interface:
            nodeinfo = Gio.DBusNodeInfo.new_for_xml(MPRIS_INTROSPECTION)
            self.root_interface = nodeinfo.interfaces[0]
            self.player_interface = nodeinfo.interfaces[1]
        self.exaile = exaile
        self.connection = None
        self.object = None
        self.registrations = []
        Gio.bus_own_name(
            Gio.BusType.SESSION,
            'org.mpris.MediaPlayer2.exaile',
            Gio.BusNameOwnerFlags.NONE,
            self._on_bus_acquired,
            None,
            None,
        )

    def _on_bus_acquired(self, connection, name):
        self.connection = connection
        self.object = obj = mprisobject.MprisObject(self.exaile, connection)
        helper = dbushelper.DBusHelper(obj)
        self.registrations.append(
            connection.register_object(
                '/org/mpris/MediaPlayer2',
                self.root_interface,
                helper.method_call,
                helper.get_property,
                helper.set_property,
            )
        )
        self.registrations.append(
            connection.register_object(
                '/org/mpris/MediaPlayer2',
                self.player_interface,
                helper.method_call,
                helper.get_property,
                helper.set_property,
            )
        )

    def disconnect(self):
        # XXX: Should probably cancel bus_own_name if it's still running.
        # FIXME: disable+enable doesn't work
        connection, self.connection = self.connection, None
        registrations = self.registrations
        while registrations:
            connection.unregister_object(registrations.pop())
        if connection:
            connection.close_sync()

    def teardown(self):
        obj = self.object
        if obj:
            obj.teardown()

    def __del__(self):
        self.disconnect()



from xl.nls import gettext as _
from xlgui.widgets import menu


class JamendoMenu(menu.Menu):
    def __init__(self, parent):
        menu.Menu.__init__(self, parent)

        self.add_item(
            menu.simple_menu_item(
                'append',
                [],
                _('Append to Current'),
                'gtk-add',
                callback=lambda *args: parent.add_to_playlist(),
            )
        )

        # self.add_item(menu.simple_menu_item('download', ['append'], _('Download to Library'), 'gtk-save',
        #                        callback=lambda *args: parent.add_to_playlist()))



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2020 Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
    This is the Native Language Support module.  It basically allows us to
    code in a gettext fashion without a hard depend on gettext itself.
"""

import locale
import os.path
import sys
from typing import Optional

try:
    import gettext as gettextmod
except ImportError:
    gettextmod = None


def _get_locale_path() -> Optional[str]:
    # This is the same as xl.xdg.exaile_path. We don't want to import xl.xdg
    # here because it's a heavy import (it pulls in GLib).
    exaile_path = os.environ['EXAILE_DIR']

    # Check if running from source dir
    locale_path = os.path.join(exaile_path, 'build', 'locale')
    if os.path.exists(locale_path):  # Equivalent to xl.xdg.local_hack
        return locale_path

    # Check if installed
    lib_suffix = os.path.join('lib', 'exaile')
    if exaile_path.endswith(lib_suffix):
        prefix = exaile_path[: -len(lib_suffix)]
        locale_path = os.path.join(prefix, 'share', 'locale')
        if os.path.exists(locale_path):
            return locale_path

    return None


def _setup_locale() -> None:
    try:
        # Required for Gtk.Builder messages
        locale.textdomain('exaile')
    except AttributeError:  # E.g. Windows
        pass

    # Required for dynamically added messages
    gettextmod.textdomain('exaile')

    locale_path = _get_locale_path()
    if locale_path is not None:
        try:
            locale.bindtextdomain('exaile', locale_path)
        except AttributeError:  # E.g. Windows
            pass
        gettextmod.bindtextdomain('exaile', locale_path)


try:
    # Set to user default, gracefully fallback on C otherwise
    locale.setlocale(locale.LC_ALL, '')
except locale.Error as e:
    # Error message copied from bzr
    sys.stderr.write(
        'exaile: Warning: %s\n'
        '  Exaile could not set the application locale, this\n'
        '  may cause language-specific problems. To investigate this\n'
        '  issue, look at the output of the locale tool.\n' % e
    )


if gettextmod:
    _setup_locale()
    gettext = gettextmod.gettext
    ngettext = gettextmod.ngettext
else:
    # gettext is not available; provide dummy functions instead

    def gettext(text: str) -> str:
        return text

    def ngettext(singular: str, plural: str, n: float) -> str:
        if n == 1:
            return singular
        else:
            return plural



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from typing import Callable, Iterable, List, Optional, TypeVar

from gi.repository import Gio
from gi.repository import GLib

from xl import metadata
from xl.trax.track import Track
from xl.trax.search import search_tracks, TracksMatcher

T = TypeVar('T')


def is_valid_track(location):
    """
    Returns whether the file at the given location is a valid track

    :param location: the location to check
    :type location: string
    :returns: whether the file is a valid track
    :rtype: boolean
    """
    extension = Gio.File.new_for_uri(location).get_basename().split(".")[-1]
    return extension.lower() in metadata.formats


def get_uris_from_tracks(tracks):
    """
    Returns all URIs for tracks

    :param tracks: the tracks to retrieve the URIs from
    :type tracks: list of :class:`xl.trax.Track`
    :returns: the uris
    :rtype: list of string
    """
    return [track.get_loc_for_io() for track in tracks]


def get_tracks_from_uri(uri):
    """
    Returns all valid tracks located at uri

    :param uri: the uri to retrieve the tracks from
    :type uri: string
    :returns: the retrieved tracks
    :rtype: list of :class:`xl.trax.Track`
    """
    tracks = []

    gloc = Gio.File.new_for_uri(uri)

    # don't do advanced checking on streaming-type uris as it can fail or
    # otherwise be terribly slow.
    # TODO: move uri definition somewhere more common for easy reuse?
    if gloc.get_uri_scheme() in ('http', 'mms', 'cdda'):
        return [Track(uri)]

    try:
        file_type = gloc.query_info(
            "standard::type", Gio.FileQueryInfoFlags.NONE, None
        ).get_file_type()
    except GLib.Error:  # E.g. cdda
        file_type = None
    if file_type == Gio.FileType.DIRECTORY:
        # TODO: refactor Library so we dont need the collection obj
        from xl.collection import Library, Collection

        tracks = Collection('scanner')
        lib = Library(uri)
        lib.set_collection(tracks)
        lib.rescan()
        tracks = tracks.get_tracks()
    else:
        tracks = [Track(uri)]
    return tracks


def sort_tracks(
    fields: Iterable[str],
    items: Iterable[T],
    trackfunc: Optional[Callable[[T], Track]] = None,
    reverse: bool = False,
    artist_compilations: bool = False,
) -> List[T]:
    """
    Sorts tracks.

    :param fields: tag names to sort by
    :param items: the tracks to sort,
        alternatively use *trackfunc*
    :param trackfunc: function to get a *Track*
        from an item in the *items* iterable
    :param reverse: whether to sort in reversed order
    """
    fields = list(fields)  # we need the index method
    if trackfunc is None:
        trackfunc = lambda tr: tr
    keyfunc = lambda tr: [
        trackfunc(tr).get_tag_sort(field, artist_compilations=artist_compilations)
        for field in fields
    ]
    return sorted(items, key=keyfunc, reverse=reverse)


def sort_result_tracks(fields, trackiter, reverse=False, artist_compilations=False):
    """
    Sorts SearchResultTracks, ie. the output from a search.

    Same params as sort_tracks.
    """
    return sort_tracks(
        fields, trackiter, lambda tr: tr.track, reverse, artist_compilations
    )


def get_rating_from_tracks(tracks):
    """
    Returns the common rating for all tracks or
    simply 0 if not all tracks have the same
    rating. Same goes if the amount of tracks
    is 0 or more than the internal limit.

    :param tracks: the tracks to retrieve the rating from
    :type tracks: iterable
    """
    if len(tracks) < 1:
        return 0

    # TODO: still needed?
    #    if len(tracks) > settings.get_option('rating/tracks_limit', 100):
    #        return 0

    rating = tracks[0].get_rating()

    for track in tracks:
        if track.get_rating() != rating:
            return 0

    return rating


def get_album_tracks(tracksiter, track, artist_compilations=False):
    """
    Get any tracks from the given iterable that appear to be part of
    the same album as track. If track is in the iterable, it will be
    included in the result. If there is insufficient information to
    determine the album, the empty list will be returned, even if the
    track is in the iterable.
    """
    if not all(track.get_tag_raw(t) for t in ['artist', 'album']):
        return []
    matchers = [
        TracksMatcher(track.get_tag_search(t, artist_compilations=artist_compilations))
        for t in ('artist', 'album')
    ]
    return (r.track for r in search_tracks(tracksiter, matchers))


def recursive_tracks_from_file(gfile: Gio.File) -> Iterable[Track]:
    """
    Get recursive tracks from Gio.File
    If it's a directory, expands
    Gets only valid tracks
    """
    ftype = gfile.query_info(
        'standard::type', Gio.FileQueryInfoFlags.NONE, None
    ).get_file_type()
    if ftype == Gio.FileType.DIRECTORY:
        file_infos = gfile.enumerate_children(
            'standard::name', Gio.FileQueryInfoFlags.NONE, None
        )
        files = (gfile.get_child(fi.get_name()) for fi in file_infos)
        for sub_gfile in files:
            for i in recursive_tracks_from_file(sub_gfile):
                yield i
    else:
        uri = gfile.get_uri()
        if is_valid_track(uri):
            yield Track(uri)



# Copyright (C) 2012 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import os

from xl.nls import gettext as _
from xlgui.preferences import playback

name = _('Preview Device')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'previewprefs.ui')
icon = 'media-playback-start'


def __autoconfig():
    """
    If the user hasn't used our plugin before, then try to
    autoconfig their audio settings to use a different audio
    device if possible..

    TODO: It would be cool if we could notify the user that
    a new device was plugged in...
    """

    from xl import settings

    if settings.get_option('preview_device/audiosink', None) is not None:
        return

    sink = settings.get_option('player/audiosink', None)
    if sink is None:
        return

    settings.set_option('preview_device/audiosink', sink)

    main_device = settings.get_option('player/audiosink_device', None)
    if main_device is None:
        return

    # TODO: If we ever add another engine, need to make sure that
    #       gstreamer-specific stuff doesn't accidentally get loaded
    from xl.player.gst.sink import get_devices

    # pick the first one that isn't the main device and isn't 'Auto'
    # -> if the main device is '', then it's auto. So... we actually
    # iterate backwards, assuming that the ordering matters
    for _unused, device_id, _unused in reversed(list(get_devices())):
        if device_id != main_device and name != 'auto':
            settings.set_option('preview_device/audiosink_device', device_id)
            break


__autoconfig()


class PreviewDeviceEnginePreference(playback.EnginePreference):
    name = 'preview_device/engine'


class PreviewDeviceAudioSinkPreference(playback.AudioSinkPreference):
    name = 'preview_device/audiosink'


class PreviewDeviceCustomAudioSinkPreference(playback.CustomAudioSinkPreference):
    name = 'preview_device/custom_sink_pipe'
    condition_preference_name = 'preview_device/audiosink'


class PreviewDeviceSelectDeviceForSinkPreference(
    playback.SelectDeviceForSinkPreference
):
    name = 'preview_device/audiosink_device'
    condition_preference_name = 'preview_device/audiosink'


class PreviewDeviceUserFadeTogglePreference(playback.UserFadeTogglePreference):
    name = 'preview_device/user_fade_enabled'
    condition_preference_name = 'preview_device/engine'


class PreviewDeviceUserFadeDurationPreference(playback.UserFadeDurationPreference):
    name = 'preview_device/user_fade'
    condition_preference_name = 'preview_device/engine'


class PreviewDeviceCrossFadingPreference(playback.CrossfadingPreference):
    default = False
    name = 'preview_device/crossfading'
    condition_preference_name = 'preview_device/engine'


class PreviewDeviceCrossfadeDurationPreference(playback.CrossfadeDurationPreference):
    default = 1000
    name = 'preview_device/crossfade_duration'
    condition_preference_name = 'preview_device/engine'



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import os
from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _('ReplayGain')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "replaygainprefs_pane.ui")


class AlbumModePreference(widgets.CheckPreference):
    default = True
    name = 'replaygain/album-mode'


class ClippingProtectionPreference(widgets.CheckPreference):
    default = True
    name = 'replaygain/clipping-protection'


class PreAmpPreference(widgets.SpinPreference):
    default = 0
    name = 'replaygain/pre-amp'


class FallbackGainPreference(widgets.SpinPreference):
    default = 0
    name = 'replaygain/fallback-gain'



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gio


def migrate(db, pdata, oldversion, newversion):
    for k in (x for x in pdata.keys() if x.startswith("tracks-")):
        p = pdata[k]
        tags = p[0]
        try:
            loc = tags['__loc']
        except KeyError:
            continue
        if not loc or not loc.startswith("file://"):
            continue
        loc = loc[7:]
        gloc = Gio.File.new_for_uri(loc)
        uri = gloc.get_uri()
        tags['__loc'] = uri
        pdata[k] = (tags, p[1], p[2])

    if '_serial_libraries' in pdata:
        libs = pdata['_serial_libraries']
        for l in libs:
            l['location'] = Gio.File.new_for_uri(l['location']).get_uri()
        pdata['_serial_libraries'] = libs

    pdata['_dbversion'] = newversion
    pdata.sync()



# Copyright (C) 2011  Mathias Brodala <info@noctus.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from gi.repository import Gtk, Pango, GdkPixbuf

import logging
import os.path
from threading import Thread, Timer

import pylast

from xl import common, event, player, providers, settings
from xl.nls import gettext as _
from xlgui import icons
from xlgui.widgets.menu import MenuItem
from xlgui.widgets.playlist_columns import Column, ColumnMenuItem

from . import lastfmlove_preferences
from .cellrenderertoggleimage import CellRendererToggleImage

LASTFMLOVER = None
logger = logging.getLogger(__name__)

basedir = os.path.dirname(os.path.realpath(__file__))
icons.MANAGER.add_icon_name_from_directory('love', os.path.join(basedir, 'icons'))
icons.MANAGER.add_icon_name_from_directory(
    'send-receive', os.path.join(basedir, 'icons')
)


class LastFMPlugin:
    def enable(self, exaile):
        """
        Handles the deferred enable call
        """
        self.__lastfmlover = LastFMLover()

    def disable(self, exaile):
        """
        Disables the desktop cover plugin
        """
        self.__lastfmlover.destroy()
        self.__lastfmlover = None

    def get_preferences_pane(self):
        return lastfmlove_preferences


plugin_class = LastFMPlugin


class LoveColumn(Column):
    name = 'loved'
    display = _('Loved')
    menu_title = _('Last.fm Loved')
    size = 50
    renderer = CellRendererToggleImage
    datatype = bool
    dataproperty = 'active'
    last_fm_lover = None

    def __init__(self, *args):
        Column.__init__(self, *args)

        self.model = None

        pixbuf = icons.MANAGER.pixbuf_from_icon_name('love')
        pixbuf = self._set_icon_size(pixbuf)

        self.cellrenderer.props.pixbuf = pixbuf
        self.cellrenderer.connect('toggled', self.on_toggled)

    def data_func(self, column, cellrenderer, model, iter, user_data):
        """
        Displays the loved state
        """
        track = model.get_value(iter, 0)
        lastfm_track = pylast.Track(
            track.get_tag_display('artist'),
            track.get_tag_display('title'),
            self.last_fm_lover.network,
        )
        cellrenderer.props.active = lastfm_track in self.last_fm_lover.loved_tracks

        if self.last_fm_lover.network is None:
            cellrenderer.props.sensitive = False
            cellrenderer.props.render_prelit = False
        else:
            cellrenderer.props.sensitive = True
            cellrenderer.props.render_prelit = True

        self.model = model

    def on_toggled(self, cellrenderer, path):
        """
        Loves or unloves the selected track
        """
        if cellrenderer.props.sensitive and self.last_fm_lover.network is not None:
            track = self.model.get_value(self.model.get_iter(path), 0)
            self.last_fm_lover.toggle_loved(track)

    def _set_icon_size(self, pixbuf):
        if pixbuf is None:
            return pixbuf

        font = settings.get_option('gui/playlist_font', None)
        if font is not None:
            # get default font
            default = float(Gtk.Widget.get_default_style().font_desc.get_size())
            new_font = Pango.FontDescription(font).get_size()

            # scale pixbuf accordingly
            t = GdkPixbuf.InterpType.BILINEAR
            s = max(int(pixbuf.get_width() * (new_font / default)), 1)
            pixbuf = pixbuf.scale_simple(s, s, t)

        return pixbuf

    def on_option_set(self, typ, obj, data):
        """
        Refresh playlist love icon size on playlist font option set
        """
        if data == "gui/playlist_font":
            self.cellrenderer.props.pixbuf = self._set_icon_size(
                self.cellrenderer.props.pixbuf
            )


class LoveMenuItem(MenuItem):
    """
    A menu item representing the loved state of a
    track and allowing for loving and unloving it
    """

    def __init__(self, last_fm_lover, after, get_tracks_function=None):
        MenuItem.__init__(self, 'loved', None, after)
        LoveColumn.last_fm_lover = last_fm_lover
        self.__lastfmlover = last_fm_lover
        self.get_tracks_function = get_tracks_function

    def factory(self, menu, parent, context):
        """
        Sets up the menu item
        """
        item = Gtk.ImageMenuItem.new_with_mnemonic(_('_Love This Track'))
        item.set_image(Gtk.Image.new_from_icon_name('love', Gtk.IconSize.MENU))

        if self.get_tracks_function is not None:
            tracks = self.get_tracks_function()
            empty = len(tracks) == 0
        else:
            empty = context.get('selection-empty', True)
            if not empty:
                tracks = context.get('selected-tracks', [])

        if not empty and self.__lastfmlover.network is not None:
            # We only care about the first track
            track = tracks[0]
            lastfm_track = pylast.Track(
                track.get_tag_display('artist'),
                track.get_tag_display('title'),
                self.__lastfmlover.network,
            )

            if lastfm_track in self.__lastfmlover.loved_tracks:
                item.set_label(_('Unlove This Track'))

            item.connect('activate', self.on_activate, track)
        else:
            item.set_sensitive(False)

        return item

    def on_activate(self, menuitem, track):
        """
        Loves or unloves the selected track
        """
        self.__lastfmlover.toggle_loved(track)


class LastFMLover:
    """
    Allows for retrieval and setting
    of loved tracks via Last.fm
    """

    def __init__(self):
        """
        Sets up the connection to Last.fm
        as well as the graphical interface
        """
        self.network = None
        self.user = None
        self.loved_tracks = []
        self.timer = None
        self.column_menu_item = ColumnMenuItem(column=LoveColumn, after=['__rating'])
        self.menu_item = LoveMenuItem(self, after=['rating'])

        def get_tracks_function():
            """
            Drop in replacement for menu item context
            to retrieve the currently playing track
            """
            current_track = player.PLAYER.current

            if current_track is not None:
                return [current_track]

            return []

        self.tray_menu_item = LoveMenuItem(
            self, after=['rating'], get_tracks_function=get_tracks_function
        )

        self.setup_network()

        providers.register('playlist-columns', LoveColumn)
        providers.register('playlist-columns-menu', self.column_menu_item)
        providers.register('playlist-context-menu', self.menu_item)
        providers.register('tray-icon-context', self.tray_menu_item)

        event.add_ui_callback(self.on_option_set, 'plugin_lastfmlove_option_set')

    def destroy(self):
        """
        Cleanups
        """
        event.remove_callback(self.on_option_set, 'plugin_lastfmlove_option_set')

        providers.unregister('tray-icon-context', self.tray_menu_item)
        providers.unregister('playlist-context-menu', self.menu_item)
        providers.unregister('playlist-columns-menu', self.column_menu_item)
        providers.unregister('playlist-columns', LoveColumn)

        if self.timer is not None and self.timer.is_alive():
            self.timer.cancel()

    def setup_network(self):
        """
        Tries to set up the network, retrieve the user
        and the initial list of loved tracks
        """
        try:
            self.network = pylast.LastFMNetwork(
                api_key=settings.get_option('plugin/lastfmlove/api_key', 'K'),
                api_secret=settings.get_option('plugin/lastfmlove/api_secret', 'S'),
                username=settings.get_option('plugin/ascrobbler/user', ''),
                password_hash=settings.get_option('plugin/ascrobbler/password', ''),
            )
            self.user = self.network.get_user(self.network.username)
        except Exception as e:
            self.network = None
            self.user = None

            if self.timer is not None and self.timer.is_alive():
                self.timer.cancel()

            logger.warning('Error while connecting to Last.fm network: {0}'.format(e))
        else:
            thread = Thread(target=self.get_loved_tracks)
            thread.daemon = True
            thread.start()

            logger.info('Connection to Last.fm network successful')

    def restart_timer(self):
        """
        Restarts the timer which starts the retrieval of tracks
        """
        if self.timer is not None and self.timer.is_alive():
            self.timer.cancel()

        self.timer = Timer(
            settings.get_option('plugin/lastfmlove/refresh_interval', 3600),
            self.get_loved_tracks,
        )
        self.timer.daemon = True
        self.timer.start()

    def get_loved_tracks(self):
        """
        Updates the list of loved tracks
        """
        logger.debug('Retrieving list of loved tracks...')

        try:
            tracks = self.user.get_loved_tracks(limit=None)
            # Unwrap pylast.Track from pylast.LovedTrack
            self.loved_tracks = [l.track for l in tracks]
        except Exception as e:
            logger.warning('Failed to retrieve list of loved tracks: {0}'.format(e))

        self.restart_timer()

    def toggle_loved(self, track):
        """
        Toggles the loved state of a track

        :param track: the track to love/unlove
        :type track: `xl.trax.Track`
        """
        lastfm_track = pylast.Track(
            track.get_tag_display('artist'),
            track.get_tag_display('title'),
            self.network,
        )

        if lastfm_track in self.loved_tracks:
            self.unlove_track(lastfm_track)
        else:
            self.love_track(lastfm_track)

    @common.threaded
    def love_track(self, track):
        """
        Loves a track

        :param track: the track to love
        :type track: `pylast.Track`
        """
        try:
            track.love()
        except Exception as e:
            logger.warning('Error while loving track {0}: {1}'.format(track, e))
        else:
            self.loved_tracks.append(track)
            logger.info('Loved track {0}'.format(track))

    @common.threaded
    def unlove_track(self, track):
        """
        Unloves a track

        :param track: the track to unlove
        :type track: `pylast.Track`
        """
        try:
            track.unlove()
        except Exception as e:
            logger.warning('Error while unloving track {0}: {1}'.format(track, e))
        else:
            self.loved_tracks.remove(track)
            logger.info('Unloved track {0}'.format(track))

    def on_option_set(self, event, settings, option):
        """
        Takes action upon setting changes
        """
        if option in ('plugin/lastfmlove/api_key', 'plugin/lastfmlove/api_secret'):
            self.setup_network()
        elif option == 'plugin/lastfmlove/refresh_interval':
            self.restart_timer()



# This file contains some code to test the DAAPClient as stand-alone application.

import sys
import logging

from .client import DAAPClient

log = logging.getLogger(__name__)


def main():
    connection = DAAPClient()

    if len(sys.argv) > 1:
        host = sys.argv[1]
    else:
        host = "localhost"
    if len(sys.argv) > 2:
        port = sys.argv[2]
    else:
        port = 3689

    logging.basicConfig(
        level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s'
    )

    try:
        # do everything in a big try, so we can disconnect at the end
        connection.connect(host, port)

        # auth isn't supported yet. Just log in
        session = connection.login()

        library = session.library()
        log.debug("Library name is `%r`", library.name)

        tracks = library.tracks()

        # demo - save the first track to disk
        # print("Saving %s by %s to disk as 'track.mp3'"%(tracks[0].name, tracks[0].artist))
        # tracks[0].save("track.mp3")
        if len(tracks) > 0:
            tracks[0].atom.printTree()
        else:
            print('No Tracks')
        session.update()
        print(session.revision)

    finally:
        # this here, so we logout even if there's an error somewhere,
        # or itunes will eventually refuse more connections.
        print("--------------")
        try:
            session.logout()
        except Exception:
            pass


if __name__ == '__main__':
    main()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
    Central storage of application and user settings
"""

import ast
from configparser import RawConfigParser, NoSectionError, NoOptionError
import logging
import os
import sys
from typing import Any, ClassVar

logger = logging.getLogger(__name__)

from xl import event, xdg
from xl.common import VersionError, glib_wait, glib_wait_seconds
from xl.nls import gettext as _

MANAGER = None


class SettingsManager(RawConfigParser):
    """
    Manages Exaile's settings
    """

    VERSION: ClassVar[int] = 2

    _last_serial: ClassVar[int] = 0

    # xl.common.glib_wait needs instances of this class to be hashable to use
    # as key in its WeakKeyDictionary. We simply use an increasing serial
    # number as this hash.
    _serial: int

    def __init__(self, location=None, default_location=None):
        """
        Sets up the settings manager. Expects a location
        to a file where settings will be stored. Also sets up
        periodic saves to disk.

        :param location: the location to save the settings to,
            settings will never be stored if this is None
        :type location: str or None
        :param default_location: the default location to
            initialize settings from
        """
        RawConfigParser.__init__(self)

        self.location = location
        self._saving = False
        self._dirty = False

        self._serial = self.__class__._last_serial = self.__class__._last_serial + 1

        if default_location is not None:
            try:
                self.read(default_location)
            except Exception:
                pass

        if location is not None:
            try:
                self.read(self.location) or self.read(
                    self.location + ".new"
                ) or self.read(self.location + ".old")
            except Exception:
                pass

        version = self.get_option('settings/version')
        if version and version > self.VERSION:
            raise VersionError(_('Settings version is newer than current.'))
        if version != self.VERSION:
            self.set_option('settings/version', self.VERSION)

        # save settings every 30 seconds
        if location is not None:
            self._timeout_save()

    def __hash__(self):
        return self._serial

    @glib_wait_seconds(30)
    def _timeout_save(self):
        """Save every 30 seconds"""
        self.save()
        return True

    def copy_settings(self, settings):
        """
        Copies one all of the settings contained
        in this instance to another

        :param settings: the settings object to copy to
        :type settings: :class:`xl.settings.SettingsManager`
        """
        for section in self.sections():
            for (key, value) in self.items(section):
                settings._set_direct('%s/%s' % (section, key), value)

    def clone(self):
        """
        Creates a copy of this settings object
        """
        settings = SettingsManager(None)
        self.copy_settings(settings)
        return settings

    def set_option(self, option, value, save=True):
        """
        Set an option (in ``section/key`` syntax) to the specified value

        :param option: the full path to an option
        :type option: string
        :param value: the value the option should be assigned
        :type value: any
        :param save: If True, cause the settings to be written to file
        """
        value = self._val_to_str(value)
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]

        try:
            self.set(section, key, value)
        except NoSectionError:
            self.add_section(section)
            self.set(section, key, value)

        self._dirty = True

        if save:
            self.delayed_save()

        section = section.replace('/', '_')

        event.log_event('option_set', self, option)
        event.log_event('%s_option_set' % section, self, option)

    def get_option(self, option: str, default: Any = None) -> Any:
        """
        Get the value of an option (in ``section/key`` syntax),
        returning *default* if the key does not exist yet

        :param option: the full path to an option
        :param default: a default value to use as fallback
        :returns: the option value or *default*
        """
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]

        try:
            value = self.get(section, key)
            value = self._str_to_val(value)
        except NoSectionError:
            value = default
        except NoOptionError:
            value = default

        return value

    def has_option(self, option):
        """
        Returns information about the existence
        of a particular option

        :param option: the option path
        :type option: string
        :returns: whether the option exists or not
        :rtype: bool
        """
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]

        return RawConfigParser.has_option(self, section, key)

    def remove_option(self, option):
        """
        Removes an option (in ``section/key`` syntax),
        thus will not be saved anymore

        :param option: the option path
        :type option: string
        """
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]

        RawConfigParser.remove_option(self, section, key)

    def _set_direct(self, option, value):
        """
        Sets the option directly to the value,
        only for use in copying settings.

        :param option: the option path
        :type option: string
        :param value: the value to set
        :type value: any
        """
        splitvals = option.split('/')
        section, key = "/".join(splitvals[:-1]), splitvals[-1]

        try:
            self.set(section, key, value)
        except NoSectionError:
            self.add_section(section)
            self.set(section, key, value)

        event.log_event('option_set', self, option)

    def _val_to_str(self, value):
        """
        Turns a value of some type into a string so it
        can be a configuration value.
        """
        for kind, type_ in (
            # bool is subclass of int so it must appear earlier
            ('B', bool),
            ('I', int),
            ('F', float),
            ('L', list),
            ('D', dict),
        ):
            if isinstance(value, type_):
                return '%s: %r' % (kind, value)
        if isinstance(value, str):
            return 'S: %s' % value  # Not quoted, hence %s

        raise ValueError(
            "Don't know how to store setting %r of type %s" % (value, type(value))
        )

    def _str_to_val(self, configstr):
        """
        Convert setting strings back to normal values.
        """
        try:
            kind, value = configstr.split(': ', 1)
        except ValueError:
            return ''

        if kind == 'B':  # Must appear before I
            return value == 'True'
        if kind in 'SU':  # U is for backwards compatibility (Python 2 unicode)
            return value
        try:
            if kind == 'I':
                return int(value)
            if kind == 'F':
                return float(value)
            for kind in 'LD':
                return ast.literal_eval(value)
        except Exception:
            logger.exception("Failed decoding value %r of kind %s", value, kind)
            return value

        raise ValueError("Unknown type of setting on: %s", configstr)

    @glib_wait(500)
    def delayed_save(self):
        '''Save options after a delay, waiting for multiple saves to accumulate'''
        self.save()

    def save(self):
        """
        Save the settings to disk
        """
        if self.location is None:
            logger.debug("Save requested but not saving settings, " "location is None")
            return

        if self._saving or not self._dirty:
            return

        self._saving = True

        logger.debug("Saving settings...")

        with open(self.location + ".new", 'w') as f:
            self.write(f)

            try:
                # make it readable by current user only, to protect private data
                os.fchmod(f.fileno(), 384)
            except Exception:
                pass  # fail gracefully, eg if on windows

            f.flush()

        try:
            os.rename(self.location, self.location + ".old")
        except Exception:
            pass  # if it doesn'texist we don't care

        os.rename(self.location + ".new", self.location)

        try:
            os.remove(self.location + ".old")
        except Exception:
            pass

        self._saving = False
        self._dirty = False


location = xdg.get_config_dir()


# Provide a mechanism for setting up default settings for different platforms
if sys.platform == 'win32':
    __settings_file = 'settings-win32.ini'
elif sys.platform == 'darwin':
    __settings_file = 'settings-osx.ini'
else:
    __settings_file = 'settings.ini'


MANAGER = SettingsManager(
    os.path.join(location, "settings.ini"), xdg.get_config_path("settings.ini")
)

get_option = MANAGER.get_option
set_option = MANAGER.set_option

# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""
    A generic framework for service providers, recommended to be used
    whenever there are multiple ways of accomplishing a task or multiple
    sources can offer the required data.
"""

from xl import event
import logging

logger = logging.getLogger(__name__)


class ProviderManager:
    """
    The overall manager for services and providers for them
    """

    def __init__(self):
        self.services = {}

    def register_provider(self, servicename, provider, target=None):
        """
        Registers a provider for a service. The provider object is used
        by consumers of the service.

        Services can be targeted for a specific use. For example, if you
        have a widget that uses a service 'foo', if your object can perform
        a service only for a specific type of widget, then target would be
        set to the widget type.

        If you had a service that could perform 'foo' for all widgets, then
        target would be set to None, and all widgets could use your service.

        It is intended that most services should set target to None, with
        some narrow exceptions.

        :param servicename: the name of the service [string]
        :type servicename: string
        :param provider: the object that is the provider [object]
        :type provider: object
        :param target: a specific target for the service [object]
        :type target: object
        """
        service = self.services.setdefault(servicename, {})
        providers = service.setdefault(target, [])
        if provider not in providers:
            providers.append(provider)
            logger.debug(
                "Provider %(provider)s registered for service %(service)s "
                "with target %(target)s"
                % {'provider': provider.name, 'service': servicename, 'target': target}
            )
            event.log_event("%s_provider_added" % servicename, self, (provider, target))

    def unregister_provider(self, servicename, provider, target=None):
        """
        Unregisters a provider.

        :param servicename: the name of the service
        :type servicename: string
        :param provider: the provider to be removed
        :type provider: object
        :param target: a specific target for the service [object]
        :type target: object
        """
        if servicename not in self.services:
            return
        try:
            service = self.services[servicename]
            if provider in service[target]:
                service[target].remove(provider)
                logger.debug(
                    "Provider %(provider)s unregistered from "
                    "service %(service)s with target %(target)s"
                    % {
                        'provider': provider.name,
                        'service': servicename,
                        'target': target,
                    }
                )
                event.log_event(
                    "%s_provider_removed" % servicename, self, (provider, target)
                )
                if not service[target]:  # no values for target key then del it
                    del service[target]
        except KeyError:
            return

    def get_providers(self, servicename, target=None):
        """
        Returns a list of providers for the specified servicename.

        This will return providers targeted for a specific target AND
        providers not targeted towards any particular target.

        :param servicename: the service name to get providers for
        :type servicename: string
        :param target: the target of the service
        :type target: object
        :returns: list of providers
        :rtype: list of objects
        """
        try:
            service = self.services[servicename]
        except KeyError:
            return []

        try:
            generic = service[None]
        except KeyError:
            generic = []

        if target is None:
            return generic[:]

        try:
            specific = service[target]
        except KeyError:
            specific = []

        return specific + generic

    def get_provider(self, servicename, providername, target=None):
        """
        Returns a single identified provider

        This will return a provider either targeted for the specific
        target or a provider not targeted towards any particular target.

        :param servicename: The service name to get the provider for
        :type servicename: string
        :param providername: The provider name to identify the provider
        :type providername: string
        :param target: the target of the service
        :type target: object
        :returns: a provider or None
        :rtype: object
        """
        for provider in self.get_providers(servicename, target):
            if provider.name == providername:
                return provider

        return None


MANAGER = ProviderManager()
register = MANAGER.register_provider
unregister = MANAGER.unregister_provider
get = MANAGER.get_providers
get_provider = MANAGER.get_provider


class ProviderHandler:
    """
    Base class to handle providers
    for one specific service including
    notification about (un)registration
    """

    def __init__(self, servicename, target=None, simple_init=False):
        """
        Target is the object that the service is being performed for.
        Often, if the service is truly global and it doesn't make sense
        to target a service at a particular consumer, it can be None.

        :param servicename: the name of the service to handle
        :type servicename: string
        :param target: the target for a provided service. Generally,
                       this will be the object that uses the service
        :type target: string
        :param simple_init: call on_provider_added for every element
                            already registered on instantiation.
        """
        self.servicename = servicename
        self.target = target
        if simple_init:
            for provider in MANAGER.get_providers(servicename, target):
                self.on_provider_added(provider)
        event.add_ui_callback(self._add_callback, "%s_provider_added" % servicename)
        event.add_ui_callback(
            self._remove_callback, "%s_provider_removed" % servicename
        )

    def _add_callback(self, name, obj, ptuple):
        """
        Mediator to call actual callback
        for added providers
        """
        provider, target = ptuple
        if target is None or target is self.target:
            self.on_provider_added(provider)

    def on_provider_added(self, provider):
        """
        Called when a new provider is added

        :param provider: the new provider
        :type provider: object
        """
        pass  # for overriding

    def _remove_callback(self, name, obj, ptuple):
        """
        Mediator to call actual callback
        for removed providers
        """
        provider, target = ptuple
        if target is None or target is self.target:
            self.on_provider_removed(provider)

    def on_provider_removed(self, provider):
        """
        Called when a provider is removed

        :param provider: the removed provider
        :type provider: object
        """
        pass  # for overriding

    def get_providers(self):
        """
        Returns a list of providers for this service

        :returns: list of providers
        :rtype: list of objects
        """
        return MANAGER.get_providers(self.servicename, self.target)

    def get_provider(self, providername):
        """
        Returns a provider for this service.

        :param providername: The provider name to
            identify the provider
        :type providername: string
        :returns: A provider or None
        :rtype: object
        """
        return MANAGER.get_provider(self.servicename, providername, self.target)


class MultiProviderHandler:
    """
    This is useful for listening to multiple provider types

    TODO: optimize implementation, could be better
    """

    class _ProxyProvider(ProviderHandler):
        def __init__(self, servicename, target, simple_init, parent):
            self.parent = parent
            ProviderHandler.__init__(self, servicename, target, simple_init)

        def on_provider_added(self, provider):
            self.parent.on_provider_added(provider)

        def on_provider_removed(self, provider):
            self.parent.on_provider_removed(provider)

    def __init__(self, servicenames, target=None, simple_init=False):
        self.providers = []
        for servicename in servicenames:
            self.providers.append(
                MultiProviderHandler._ProxyProvider(
                    servicename, target, simple_init, self
                )
            )

    def on_provider_added(self, provider):
        """
        Called when a new provider is added

        :param provider: the new provider
        :type provider: object
        """

    def on_provider_removed(self, provider):
        """
        Called when a provider is removed

        :param provider: the removed provider
        :type provider: object
        """

    def get_providers(self):
        """
        Returns a list of providers for this service

        :returns: list of providers
        :rtype: list of objects
        """

        providers = []
        for provider in self.providers:
            providers.extend(provider.get_providers())
        return providers


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

__all__ = ['main', 'panel', 'playlist']

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk

import logging

logger = logging.getLogger(__name__)

import os
import sys

from xl import common, player, providers, settings, version, xdg
from xl.nls import gettext as _
from xlgui import guiutil

version.register(
    "GTK+", "%s.%s.%s" % (Gtk.MAJOR_VERSION, Gtk.MINOR_VERSION, Gtk.MICRO_VERSION)
)
version.register("GTK+ theme", Gtk.Settings.get_default().props.gtk_theme_name)


def get_controller():
    return Main._main


class Main:
    """
    This is the main gui controller for exaile
    """

    _main = None

    def __init__(self, exaile):
        """
        Initializes the GUI

        @param exaile: The Exaile instance
        """
        from xlgui import icons, main, panels, tray, progress

        Gdk.set_program_class("Exaile")  # For GNOME Shell

        # https://www.freedesktop.org/wiki/Software/PulseAudio/Documentation/Developer/Clients/ApplicationProperties/
        GLib.set_application_name("Exaile")
        os.environ['PULSE_PROP_media.role'] = 'music'

        self.exaile = exaile
        self.first_removed = False
        self.tray_icon = None

        self.builder = Gtk.Builder()
        self.builder.add_from_file(xdg.get_data_path('ui', 'main.ui'))
        self.progress_box = self.builder.get_object('progress_box')
        self.progress_manager = progress.ProgressManager(self.progress_box)

        add_icon = icons.MANAGER.add_icon_name_from_directory
        images_dir = xdg.get_data_path('images')

        exaile_icon_path = add_icon('exaile', images_dir)
        Gtk.Window.set_default_icon_name('exaile')
        if xdg.local_hack:
            # PulseAudio also attaches the above name to streams. However, if
            # Exaile is not installed, any app trying to display the icon won't
            # be able to find it just by name. The following is a hack to tell
            # PA the icon file path instead of the name; this only works on
            # some clients, e.g. pavucontrol.
            os.environ['PULSE_PROP_application.icon_name'] = exaile_icon_path

        for name in (
            'exaile-pause',
            'exaile-play',
            'office-calendar',
            'extension',
            'music-library',
            'artist',
            'genre',
        ):
            add_icon(name, images_dir)
        for name in ('dynamic', 'repeat', 'shuffle'):
            add_icon('media-playlist-' + name, images_dir)

        logger.info("Loading main window...")
        self.main = main.MainWindow(self, self.builder, exaile.collection)

        if self.exaile.options.StartMinimized:
            self.main.window.iconify()

        self.play_toolbar = self.builder.get_object('play_toolbar')

        panel_notebook = self.builder.get_object('panel_notebook')
        self.panel_notebook = panels.PanelNotebook(exaile, self)

        self.device_panels = {}

        # add the device panels
        for device in self.exaile.devices.get_devices():
            if device.connected:
                self.add_device_panel(None, None, device)

        logger.info("Connecting panel events...")
        self.main._connect_panel_events()

        guiutil.gtk_widget_replace(panel_notebook, self.panel_notebook)
        self.panel_notebook.get_parent().child_set_property(
            self.panel_notebook, 'shrink', False
        )

        if settings.get_option('gui/use_tray', False):
            if tray.is_supported():
                self.tray_icon = tray.TrayIcon(self.main)
            else:
                settings.set_option('gui/use_tray', False)
                logger.warning(
                    "Tray icons are not supported on your platform. Disabling tray icon."
                )

        from xl import event

        event.add_ui_callback(self.add_device_panel, 'device_connected')
        event.add_ui_callback(self.remove_device_panel, 'device_disconnected')
        event.add_ui_callback(self.on_gui_loaded, 'gui_loaded')

        logger.info("Done loading main window...")
        Main._main = self

        if sys.platform == 'darwin':
            self._setup_osx()

    def open_uris(self, uris, play=True):
        if len(uris) > 0:
            self.open_uri(uris[0], play=play)

            for uri in uris[1:]:
                self.open_uri(uri, play=False)

    def open_uri(self, uri, play=True):
        """
        Determines the type of a uri, imports it into a playlist, and
        starts playing it
        """
        from xl import playlist, trax

        if playlist.is_valid_playlist(uri):
            try:
                playlist = playlist.import_playlist(uri)
            except playlist.InvalidPlaylistTypeError:
                pass
            else:
                self.main.playlist_container.create_tab_from_playlist(playlist)

                if play:
                    player.QUEUE.current_playlist = playlist
                    player.QUEUE.current_playlist.current_position = 0
                    player.QUEUE.play(playlist[0])
        else:
            page = self.main.get_selected_page()
            column = page.view.get_sort_column()
            reverse = False
            sort_by = common.BASE_SORT_TAGS

            if column:
                reverse = column.get_sort_order() == Gtk.SortType.DESCENDING
                sort_by = [column.name] + sort_by

            tracks = trax.get_tracks_from_uri(uri)
            tracks = trax.sort_tracks(sort_by, tracks, reverse=reverse)

            try:
                page.playlist.extend(tracks)
                page.playlist.current_position = len(page.playlist) - len(tracks)

                if play:
                    player.QUEUE.current_playlist = page.playlist
                    player.QUEUE.play(tracks[0])
            # Catch empty directories
            except IndexError:
                pass

    def show_cover_manager(self, *e):
        """
        Shows the cover manager
        """
        from xlgui.cover import CoverManager

        CoverManager(self.main.window, self.exaile.collection)

    def show_preferences(self):
        """
        Shows the preferences dialog
        """
        from xlgui.preferences import PreferencesDialog

        dialog = PreferencesDialog(self.main.window, self)
        dialog.run()

    def show_devices(self):
        from xlgui.devices import ManagerDialog

        dialog = ManagerDialog(self.main.window, self)
        dialog.run()

    def queue_manager(self, *e):
        self.main.playlist_container.show_queue()

    def collection_manager(self, *e):
        """
        Invokes the collection manager dialog
        """
        from xl.collection import Library
        from xlgui.collection import CollectionManagerDialog

        dialog = CollectionManagerDialog(self.main.window, self.exaile.collection)
        result = dialog.run()
        dialog.hide()

        if result == Gtk.ResponseType.APPLY:
            collection = self.exaile.collection
            collection.freeze_libraries()

            collection_libraries = sorted(
                (l.location, l.monitored, l.startup_scan)
                for l in collection.libraries.values()
            )
            new_libraries = sorted(dialog.get_items())

            if collection_libraries != new_libraries:
                collection_locations = [
                    location
                    for location, monitored, startup_scan in collection_libraries
                ]
                new_locations = [
                    location for location, monitored, startup_scan in new_libraries
                ]

                if collection_locations != new_locations:
                    for location in new_locations:
                        if location not in collection_locations:
                            collection.add_library(Library(location))

                    removals = []

                    for location, library in collection.libraries.items():
                        if location not in new_locations:
                            removals.append(library)

                    for removal in removals:
                        collection.remove_library(removal)

                    self.on_rescan_collection()

                for location, monitored, startup_scan in new_libraries:
                    collection.libraries[location].monitored = monitored
                    collection.libraries[location].startup_scan = startup_scan

            collection.thaw_libraries()

        dialog.destroy()

    def on_gui_loaded(self, event, object, nothing):

        # This has to be idle_add so that plugin panels can be configured
        GLib.idle_add(self.panel_notebook.on_gui_loaded)

        # Fix track info not displaying properly when resuming after a restart.
        self.main._update_track_information()

    def on_rescan_collection(self, *e):
        """
        Called when the user wishes to rescan the collection
        """
        self.rescan_collection_with_progress()

    def on_rescan_collection_forced(self, *e):
        """
        Called when the user wishes to rescan the collection slowly
        """
        self.rescan_collection_with_progress(force_update=True)

    def rescan_collection_with_progress(self, startup=False, force_update=False):

        libraries = self.exaile.collection.get_libraries()
        if not self.exaile.collection._scanning and len(libraries) > 0:
            from xl.collection import CollectionScanThread

            thread = CollectionScanThread(
                self.exaile.collection, startup_scan=startup, force_update=force_update
            )
            thread.connect('done', self.on_rescan_done)
            self.progress_manager.add_monitor(
                thread, _("Scanning collection..."), 'drive-harddisk'
            )

    def on_rescan_done(self, thread):
        """
        Called when the rescan has finished
        """
        GLib.idle_add(self.get_panel('collection').load_tree)

    def on_track_properties(self, *e):
        pl = self.main.get_selected_page()
        pl.view.show_properties_dialog()

    def get_active_panel(self):
        """
        Returns the provider object associated with the currently shown
        panel in the sidebar. May return None.
        """
        return self.panel_notebook.get_active_panel()

    def focus_panel(self, panel_name):
        """
        Focuses on a panel in the sidebar
        """
        self.panel_notebook.focus_panel(panel_name)

    def get_panel(self, panel_name):
        """
        Returns the provider object associated with a panel in the sidebar
        """
        return self.panel_notebook.panels[panel_name].panel

    def quit(self):
        """
        Quits the gui, saving anything that needs to be saved
        """

        # save open tabs
        self.main.playlist_container.save_current_tabs()

    def add_device_panel(self, type, obj, device):
        from xl.collection import CollectionScanThread
        from xlgui.panel.device import DevicePanel, FlatPlaylistDevicePanel
        import xlgui.panel

        paneltype = DevicePanel
        if hasattr(device, 'panel_type'):
            if device.panel_type == 'flatplaylist':
                paneltype = FlatPlaylistDevicePanel
            elif issubclass(device.panel_type, xlgui.panel.Panel):
                paneltype = device.panel_type

        panel = paneltype(self.main.window, self.main, device, device.get_name())

        do_sort = True
        panel.connect(
            'append-items',
            lambda _panel, items, play: self.main.on_append_items(
                items, play, sort=do_sort
            ),
        )
        panel.connect(
            'queue-items',
            lambda _panel, items: self.main.on_append_items(
                items, queue=True, sort=do_sort
            ),
        )
        panel.connect(
            'replace-items',
            lambda _panel, items: self.main.on_append_items(
                items, replace=True, sort=do_sort
            ),
        )

        self.device_panels[device.get_name()] = panel
        GLib.idle_add(providers.register, 'main-panel', panel)
        thread = CollectionScanThread(device.get_collection())
        thread.connect('done', panel.load_tree)
        self.progress_manager.add_monitor(
            thread, _("Scanning %s..." % device.name), 'drive-harddisk'
        )

    def remove_device_panel(self, type, obj, device):
        try:
            providers.unregister('main-panel', self.device_panels[device.get_name()])
        except ValueError:
            logger.debug("Couldn't remove panel for %s", device.get_name())
        del self.device_panels[device.get_name()]

    def _setup_osx(self):
        """
        Copied from Quod Libet, GPL v2 or later
        """

        from AppKit import NSObject, NSApplication
        import objc

        try:
            import gi

            gi.require_version('GtkosxApplication', '1.0')
            from gi.repository import GtkosxApplication
        except (ValueError, ImportError):
            logger.warning("importing GtkosxApplication failed, no native menus")
        else:
            osx_app = GtkosxApplication.Application()
            # self.main.setup_osx(osx_app)
            osx_app.ready()

        shared_app = NSApplication.sharedApplication()
        gtk_delegate = shared_app.delegate()

        other_self = self

        # TODO
        # Instead of quitting when the main window gets closed just hide it.
        # If the dock icon gets clicked we get
        # applicationShouldHandleReopen_hasVisibleWindows_ and show everything.
        class Delegate(NSObject):
            @objc.signature('B@:#B')
            def applicationShouldHandleReopen_hasVisibleWindows_(self, ns_app, flag):
                logger.debug("osx: handle reopen")
                # TODO
                # app.present()
                return True

            def applicationShouldTerminate_(self, sender):
                logger.debug("osx: block termination")
                other_self.main.quit()
                return False

            def applicationDockMenu_(self, sender):
                return gtk_delegate.applicationDockMenu_(sender)

            # def application_openFile_(self, sender, filename):
            #    return app.window.open_file(filename.encode("utf-8"))

        delegate = Delegate.alloc().init()
        delegate.retain()
        shared_app.setDelegate_(delegate)

        # QL shouldn't exit on window close, EF should
        # if window.get_is_persistent():
        #    window.connect(
        #        "delete-event", lambda window, event: window.hide() or True)



# -*- coding: utf-8 -*-
#
# Exaile documentation build configuration file, created by
# sphinx-quickstart on Mon Jul 31 17:15:07 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

from os.path import abspath, join, dirname

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

os.environ['EXAILE_DIR'] = abspath(join(dirname(__file__), os.pardir))
sys.path.insert(0, os.environ['EXAILE_DIR'])

# indicate that we're building docs
os.environ['EXAILE_BUILDING_DOCS'] = 'TRUE'

EXAILE_AUTHORS = [
    'Adam Olsen <arolsen@gmail.com>',
    'Johannes Sasongko <sasongko@gmail.com>',
    'Aren Olsen <reacocard@gmail.com>',
    'Mathias Brodala <info@noctus.net>',
    'Dustin Spicuzza <dustin@virtualroadside.com>',
    'Christian Stadelmann',
]

#
# Allow building on systems which don't have all dependencies installed
#

sys.path.append(abspath(dirname(__file__)))
import mocks
mocks.import_fake_modules()
mocks.fake_xl_settings()
sys.path.pop()


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.intersphinx',
              'sphinx.ext.todo',
              'sphinx.ext.coverage',
              'sphinx.ext.viewcode']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Exaile'
copyright = '2009-2018 Exaile Development Team'
author = ''.join(EXAILE_AUTHORS)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

import xl.version
version = xl.version.__version__
release = xl.version.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for HTML output ---------------------------------------------------

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from
# docs.readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
else:
    html_theme = 'default'


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: https://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
        'donate.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Exailedoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    #
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    #
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    #
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Exaile.tex', 'Exaile Documentation',
     author.replace('<', '\\textless{}').replace('>', '\\textgreater{}'),
     'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'exaile', 'Exaile Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Exaile', 'Exaile Documentation',
     author, 'Exaile', 'One line description of project.',
     'Miscellaneous'),
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/': None}



# Copyright (C) 2014 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gio

from os.path import dirname, join

from contextlib import closing

from xl.nls import gettext as _
from xl.metadata.tags import tag_data

from xlgui.widgets import menu
from .analyzer_dialog import AnalyzerDialog


class PlaylistAnalyzerPlugin:
    def __init__(self):
        self.menu_items = []
        self.dialog = None
        self._get_track_groups = None

        self.d3_loc = join(dirname(__file__), 'ext', 'd3.min.js')

    def enable(self, exaile):
        self.exaile = exaile

    def on_gui_loaded(self):

        # register menu items
        item = menu.simple_menu_item(
            'pz-run', [], _('Analyze playlists'), callback=self.on_analyze_playlists
        )
        item.register('menubar-tools-menu')
        self.menu_items.append(item)

        item = menu.simple_menu_item(
            'pz-run',
            ['export-files'],
            _('Analyze playlist'),
            callback=self.on_analyze_playlist,
        )
        item.register('playlist-panel-context-menu')
        self.menu_items.append(item)

        # -> this could have a submenu that gets filled in with all
        #    of the presets

    def on_exaile_loaded(self):
        pass

    def disable(self, exaile):

        if self.dialog is not None:
            self.dialog.destroy()
            self.dialog = None

        for menu_item in self.menu_items:
            menu_item.unregister()

    #
    # Misc
    #

    def get_track_groups(self, track):

        if self._get_track_groups is None:

            if 'grouptagger' not in self.exaile.plugins.enabled_plugins:
                raise ValueError(
                    "GroupTagger plugin must be loaded to use the GroupTagger tag"
                )

            self._get_track_groups = self.exaile.plugins.enabled_plugins[
                'grouptagger'
            ].get_track_groups

        return self._get_track_groups(track)

    #
    # Menu functions
    #

    def on_analyze_playlist(self, widget, name, parent, context):
        """
        :param parent: The PlaylistsPanel that triggered this callback
        """

        if self.dialog is None:
            self.dialog = AnalyzerDialog(
                self, parent.parent, context['selected-playlist']
            )

    def on_analyze_playlists(self, widget, name, parent, context):
        """
        :param parent: The Exaile MainWindow object
        """

        if self.dialog is None:
            self.dialog = AnalyzerDialog(self, parent.window)

    #
    # Functions to generate the analysis
    #

    def get_tag(self, track, tagname, extra):

        data = tag_data.get(tagname)

        if data is not None:
            if data.type == 'int':
                ret = track.get_tag_raw(tagname, join=True)
                if ret is not None:
                    if extra == 0:
                        return int(ret)
                    else:
                        return int(ret) - (int(ret) % extra)
                return

            if data.use_disk:
                return track.get_tag_disk(tagname)

        if tagname == '__grouptagger':
            return list(self.get_track_groups(track))

        return track.get_tag_raw(tagname, join=True)

    def generate_data(self, tracks, tagdata):

        data = []

        for track in tracks:
            if track is None:
                data.append(None)
            else:
                data.append([self.get_tag(track, tag, extra) for tag, extra in tagdata])

        return data

    def write_to_file(self, tmpl, uri, **kwargs):
        """
        Opens a template file, performs substitution, writes it to the
        output URI, and also writes d3.min.js to the output directory.

        :param tmpl: Local pathname to template file
        :param uri: URI of output file suitable for passing to Gio.File
        :param kwargs: Named parameters to substitute in template
        """

        # read the template file
        # NOTE: must be opened in non-binary mode because we treat the
        # contents as string
        with open(tmpl, 'r') as fp:
            contents = fp.read()

        try:
            contents = contents % kwargs
        except Exception:
            raise RuntimeError(
                "Format string error in template (probably has unescaped % in it)"
            )

        outfile = Gio.File.new_for_uri(uri)
        parent_dir = outfile.get_parent()
        if parent_dir:
            parent_dir = parent_dir.get_child("d3.min.js")

        with closing(
            outfile.replace(None, False, Gio.FileCreateFlags.NONE, None)
        ) as fp:
            fp.write(
                contents.encode('utf-8')
            )  # Gio.FileOutputStream.write() needs bytes argument

        # copy d3 to the destination
        # -> TODO: add checkbox to indicate whether it should write d3 there or not
        if parent_dir:
            # Open in binary mode, so that we can directly read bytes
            # and write them via Gio.FileOutputStream.write()
            with open(self.d3_loc, 'rb') as d3fp:
                with closing(
                    parent_dir.replace(None, False, Gio.FileCreateFlags.NONE, None)
                ) as pfp:
                    pfp.write(d3fp.read())


# New plugin API; requires exaile 3.4.0 or later
plugin_class = PlaylistAnalyzerPlugin



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Arunas Radzvilavicius, arunas.rv@gmail.com

import urllib.parse
from xml.etree import ElementTree
from xl import common
import logging

logger = logging.getLogger(__name__)


# TODO: The 'new' API allows many fields and generic search. Update UI!
# TODO: New API docs: https://librivox.org/api/info
search_url = 'https://librivox.org/api/feed/audiobooks/?title='


class Book:
    def __init__(self, title, rssurl, user_agent):
        self.title = title
        self.rssurl = rssurl
        self.chapters = []
        self.info = None
        self.is_loading = False
        self.xmldata = None
        self.xmltree = None
        self.loaded = False
        self.user_agent = user_agent

    def get_all(self):
        """
        Unified function for getting chapters and info at the same
        time.
        """
        if self.loaded:
            return

        try:
            self.xmldata = common.get_url_contents(self.rssurl, self.user_agent)
        except Exception:
            logger.error("LIBRIVOX: Connection error")
            return

        try:
            self.xmltree = ElementTree.XML(self.xmldata)
        except Exception:
            logger.error("LIBRIVOX: XML error")
            return

        self.chapters = []
        items = self.xmltree.findall("channel/item")
        for item in items:
            title = item.find("title").text
            link = item.find("link").text
            duration = item.find(
                "{http://www.itunes.com/dtds/podcast-1.0.dtd}duration"
            ).text
            if duration is None:
                duration = 'Unknown length'
            link = link.replace("_64kb.mp3", ".ogg")
            self.chapters.append([title + " " + "(" + duration + ")", link])

        self.info = self.xmltree.find("channel/description")
        self.info = self.info.text
        self.loaded = True
        return


def find_books(keyword, user_agent):
    """
    Returns a list of Book instances, with unknown chapters...
    """

    # urlencode the search string
    url = search_url + urllib.parse.quote_plus(keyword)

    try:
        data = common.get_url_contents(url, user_agent)
    except Exception:
        logger.error("LIBRIVOX: connection error")
        return []

    try:
        tree = ElementTree.XML(data)
    except Exception:
        logger.error("LIBRIVOX: XML error")
        return []

    books = []

    for elem in tree:
        if elem.tag == 'error':
            logger.error('LIBRIVOX: query error: %s', elem.text)

        elif elem.tag == 'books':
            for bk in elem.findall('book'):
                title = bk.find("title").text
                rssurl = bk.find("url_rss").text
                book = Book(title, rssurl, user_agent)
                books.append(book)

    return books
